from django import forms
from django.shortcuts import redirect, render
from django.shortcuts import HttpResponse
from django.core.validators import RegexValidator
from django.contrib.auth.decorators import login_required
from django.core.exceptions import ObjectDoesNotExist

from aw.config.main import config
from aw.utils.http import ui_endpoint_wrapper_kwargs
from aw.model.job import Job
from aw.model.permission import CHOICE_PERMISSION_WRITE
from aw.model.job_credential import JobGlobalCredentials, JobUserCredentials
from aw.model.repository import Repository, CHOICES_REPOSITORY
from aw.api_endpoints.credentials import are_global_credentials
from aw.utils.permission import has_job_permission, has_credentials_permission
from aw.utils.util import get_choice_key_by_value
from aw.config.form_metadata import FORM_LABEL, FORM_HELP
from aw.views.base import choices_global_credentials, choices_repositories


class JobForm(forms.ModelForm):
    class Meta:
        model = Job
        fields = Job.form_fields
        field_order = Job.form_fields
        labels = FORM_LABEL['jobs']['manage']
        help_texts = FORM_HELP['jobs']['manage']

    credentials_default = forms.ChoiceField(
        required=False,
        widget=forms.Select,
        choices=choices_global_credentials,
        label=FORM_LABEL['jobs']['manage']['credentials_default'],
    )

    repository = forms.ChoiceField(
        required=False,
        widget=forms.Select,
        choices=choices_repositories,
        label=FORM_LABEL['jobs']['manage']['repository'],
    )

    # form not picking up regex-validator
    schedule = forms.CharField(
        max_length=Job.schedule_max_len,
        validators=[RegexValidator(
            regex=r'^(@(annually|yearly|monthly|weekly|daily|hourly))|'
                  r'(@every (\d+(s|m|h))+)|'
                  r'((((\d+,)+\d+|(\d+(\/|-|#)\d+)|\d+L?|\*(\/\d+)?|L(-\d+)?|\?|[A-Z]{3}(-[A-Z]{3})?) ?){5,7})$',
            message='The provided schedule is not in a valid cron format',
        )],
        required=False,
        help_text=Meta.help_texts['schedule'],
    )


@login_required
@ui_endpoint_wrapper_kwargs
def job_edit(request, job_id: int = None, clone: bool = False) -> HttpResponse:
    job_form = JobForm()
    form_method = 'post'
    form_api = 'job'
    job = {}

    if job_id is not None:
        try:
            job = Job.objects.get(id=job_id)

        except ObjectDoesNotExist:
            job = None

        if job is None:
            return redirect(f"/ui/jobs/manage?error=Job with ID {job_id} does not exist")

        job = job.__dict__
        if clone:
            job['name'] = f"{job['name']} - Copy"
            job.pop('id')

        else:
            if not has_job_permission(user=request.user, job=job['id'], permission_needed=CHOICE_PERMISSION_WRITE):
                return redirect(f"/ui/jobs/manage?error=Not privileged to modify the job '{job.name}'")

            form_method = 'put'
            form_api += f'/{job_id}'

    job_form_html = job_form.render(
        template_name='forms/job.html',
        context={'form': job_form, 'existing': job, 'primary_fields': Job.form_fields_primary},
    )
    return render(
        request, status=200, template_name='jobs/edit.html',
        context={'form': job_form_html, 'form_api': form_api, 'form_method': form_method}
    )


@login_required
@ui_endpoint_wrapper_kwargs
def job_clone(request, job_id: int) -> HttpResponse:
    return job_edit(request=request, job_id=job_id, clone=True)


class CredentialGlobalForm(forms.ModelForm):
    class Meta:
        model = JobGlobalCredentials
        fields = JobGlobalCredentials.form_fields
        field_order = JobGlobalCredentials.form_fields
        labels = FORM_LABEL['jobs']['credentials']
        help_texts = FORM_HELP['jobs']['credentials']

    vault_pass = forms.CharField(
        max_length=100, required=False, label=FORM_LABEL['jobs']['credentials']['vault_pass'],
    )
    become_pass = forms.CharField(
        max_length=100, required=False, label=FORM_LABEL['jobs']['credentials']['become_pass'],
    )
    connect_pass = forms.CharField(
        max_length=100, required=False, label=FORM_LABEL['jobs']['credentials']['connect_pass'],
    )
    ssh_key = forms.CharField(
        max_length=2000, required=False, label=FORM_LABEL['jobs']['credentials']['ssh_key'],
    )


class CredentialUserForm(CredentialGlobalForm):
    class Meta:
        model = JobUserCredentials
        fields = JobUserCredentials.form_fields
        field_order = JobUserCredentials.form_fields


@login_required
@ui_endpoint_wrapper_kwargs
def job_credentials_edit(request, credentials_id: int = None) -> HttpResponse:
    are_global = are_global_credentials(request)
    if are_global:
        credentials_form = CredentialGlobalForm()
    else:
        credentials_form = CredentialUserForm()

    form_method = 'post'
    form_api = 'credentials'
    credentials = {}

    if credentials_id is not None and credentials_id != 0:
        try:
            if are_global:
                credentials = JobGlobalCredentials.objects.get(id=credentials_id)

            else:
                credentials = JobUserCredentials.objects.get(id=credentials_id, user=request.user)

        except ObjectDoesNotExist:
            credentials = None

        if credentials is None:
            return redirect(f"/ui/jobs/credentials?error=Credentials with ID {credentials_id} do not exist")

        if isinstance(credentials, JobGlobalCredentials) and not has_credentials_permission(
                user=request.user,
                credentials=credentials,
                permission_needed=CHOICE_PERMISSION_WRITE,
        ):
            return redirect(
                f"/ui/jobs/credentials?error=Not privileged to modify the credentials '{credentials.name}'",
            )

        credentials = credentials.__dict__
        form_method = 'put'
        form_api += f'/{credentials_id}'

    form_api += '?global=true' if are_global else '?global=false'
    credentials_form_html = credentials_form.render(
        template_name='forms/snippet.html',
        context={'form': credentials_form, 'existing': credentials},
    )
    return render(
        request, status=200, template_name='jobs/credentials_edit.html',
        context={'form': credentials_form_html, 'form_api': form_api, 'form_method': form_method}
    )


class RepositoryGitForm(forms.ModelForm):
    class Meta:
        model = Repository
        fields = Repository.form_fields_git
        field_order = Repository.form_fields_git
        labels = FORM_LABEL['jobs']['repository']
        help_texts = FORM_HELP['jobs']['repository']

    git_credentials = forms.ChoiceField(
        required=False,
        widget=forms.Select,
        choices=choices_global_credentials,
        label=FORM_LABEL['jobs']['repository']['git_credentials'],
    )


class RepositoryStaticForm(forms.ModelForm):
    class Meta:
        model = Repository
        fields = Repository.form_fields_static
        field_order = Repository.form_fields_static
        labels = FORM_LABEL['jobs']['repository']
        help_texts = FORM_HELP['jobs']['repository']

    static_path = forms.CharField(max_length=500, initial=config['path_play'], required=False)


REPOSITORY_EDIT_FORMS = {
    'Git': RepositoryGitForm,
    'Static': RepositoryStaticForm,
}


def _job_repository_edit(request, rtype_name: str, repo_id: int = None) -> HttpResponse:
    form_method = 'post'
    form_api = 'repository'
    repository_type = get_choice_key_by_value(choices=CHOICES_REPOSITORY, find=rtype_name)
    repository = {}

    if repo_id is not None and repo_id != 0:
        # pylint: disable=R0801
        try:
            repository = Repository.objects.get(id=repo_id)

        except ObjectDoesNotExist:
            repository = None

        if repository is None:
            return redirect(f"/ui/jobs/repository?error=Repository with ID {repo_id} do not exist")

        if repository.rtype != repository_type:
            return redirect(f"/ui/jobs/repository?error=Repository with ID {repo_id} is not of type '{rtype_name}'")

        repository = repository.__dict__
        form_method = 'put'
        form_api += f'/{repo_id}'

    repository_form = REPOSITORY_EDIT_FORMS[rtype_name]()
    repository_form_html = repository_form.render(
        template_name='forms/snippet.html',
        context={'form': repository_form, 'existing': repository},
    )
    return render(
        request, status=200, template_name='jobs/repository_edit.html',
        context={
            'form': repository_form_html, 'form_api': form_api, 'form_method': form_method,
            'repository_type': repository_type,
        }
    )


@login_required
@ui_endpoint_wrapper_kwargs
def job_repository_git_edit(request, repo_id: int = None) -> HttpResponse:
    return _job_repository_edit(request=request, rtype_name='Git', repo_id=repo_id)


@login_required
@ui_endpoint_wrapper_kwargs
def job_repository_static_edit(request, repo_id: int = None) -> HttpResponse:
    return _job_repository_edit(request=request, rtype_name='Static', repo_id=repo_id)



# runs service timers in multiple threads
# base code source: https://github.com/sankalpjonn/timeloop

from threading import Thread, Event
from time import sleep
import traceback

from aw.utils.debug import log
from aw.config.hardcoded import THREAD_JOIN_TIMEOUT
from aw.model.job import Job, JobExecution
from aw.execute.play import ansible_playbook
from aw.utils.handlers import AnsibleConfigError, AnsibleRepositoryError
from aw.utils.util import get_next_cron_execution_sec, get_next_cron_execution_str, is_set


class Workload(Thread):
    FAIL_SLEEP = 5
    MAX_CONFIG_INVALID = 3

    def __init__(self, job: Job, manager, name: str, execution: JobExecution, once: bool = False, daemon: bool = True):
        Thread.__init__(self, daemon=daemon, name=name)
        self.job = job
        self.execution = execution
        self.manager = manager
        self.once = once
        self.started = False
        self.stopped = False
        self.state_stop = Event()
        self.log_name_debug = f"'{self.job.name}' (Job-ID {self.job.id}; {self.name})"
        self.log_name = f"'{self.job.name}' (Job-ID {self.job.id})"
        self.next_execution_time = None
        self.config_invalid = 0

    def stop(self) -> bool:
        log(f"Thread stopping {self.log_name_debug}", level=6)
        self.state_stop.set()

        try:
            self.join(THREAD_JOIN_TIMEOUT)
            if self.is_alive():
                log(f"Unable to join thread {self.log_name_debug}", level=5)

        except RuntimeError:
            # 'cannot join current thread'
            pass

        log(f"Stopped thread {self.log_name_debug}", level=4)
        self.started = False
        self.stopped = True
        return True

    def run_playbook(self):
        ansible_playbook(job=self.job, execution=self.execution)

    def run(self, error: bool = False) -> None:
        if self.once and self.started:
            self.stop()
            return

        if self.stopped:
            return

        if self.config_invalid >= self.MAX_CONFIG_INVALID:
            self.next_execution_time = None
            self.job.enabled = False
            self.job.save()
            log(msg=f"Disabling job {self.log_name} because of invalid config! Please fix it", level=2)
            # exit loop because it will always fail; fixing the config will replace this threat instance
            return

        if error:
            sleep(self.FAIL_SLEEP)

        self.started = True
        log(f"Entering runtime of thread {self.log_name_debug}", level=7)
        try:
            if self.once:
                self.run_playbook()
                self.stop()
                self.manager.threads.remove(self)
                return

            while not self.state_stop.is_set():
                wait_sec = get_next_cron_execution_sec(self.job.schedule)
                self.next_execution_time = get_next_cron_execution_str(schedule=self.job.schedule, wait_sec=wait_sec)
                log(
                    f"Next execution of job {self.log_name_debug} at "
                    f"{self.next_execution_time}",
                    level=7,
                )

                while not self.state_stop.wait(wait_sec):
                    if self.state_stop.is_set():
                        log(f"Exiting thread {self.log_name_debug}", level=5)
                        break

                    log(f"Starting job {self.log_name_debug}", level=5)
                    self.run_playbook()
                    break

        except (AnsibleConfigError, AnsibleRepositoryError, OSError) as err:
            self.config_invalid += 1
            log(
                msg=f"Got invalid config/environment for job {self.log_name} "
                    f"({self.config_invalid}/{self.MAX_CONFIG_INVALID}): \"{err}\"",
                level=2,
            )
            self.run(error=True)

        # pylint: disable=W0718
        except Exception as err:
            tb = traceback.format_exc(limit=256)
            log(
                msg=f"Got unexpected error while executing job {self.log_name}: \"{err}\"\n{tb}",
                level=2,
            )
            self.run(error=True)


class ThreadManager:
    def __init__(self):
        self.threads = set()
        self.thread_nr = 0
        self.stopping = False

    def start(self) -> None:
        log('Starting all threads', level=6)

        for thread in self.threads:
            if not thread.started:
                thread.start()

    def add_thread(self, job: Job, execution: JobExecution = None, once: bool = False):
        schedule = f" with schedule \"{job.schedule}\"" if is_set(job.schedule) else ''
        log(f"Adding thread for \"{job.name}\"{schedule}", level=7)
        self.thread_nr += 1
        self.threads.add(
            Workload(
                job=job,
                execution=execution,
                manager=self,
                once=once,
                name=f"Thread #{self.thread_nr}",
            )
        )

    def stop(self) -> bool:
        if self.stopping:
            return False

        self.stopping = True
        log('Stopping all threads', level=6)

        thread_list = list(self.threads)
        thread_count = len(thread_list)
        for i in range(thread_count):
            _ = thread_list[i]
            self.threads.remove(_)

        log('All threads stopped', level=3)
        return True

    def stop_thread(self, job: Job):
        log(f"Stopping thread for \"{job.name}\"", level=6)
        for thread in self.threads:
            if thread.job == job:
                if thread.started:
                    thread.stop()
                    self.threads.remove(thread)
                    log(f"Thread '{job.name}' stopped.", level=4)
                    break

    def start_thread(self, job: Job) -> None:
        for thread in self.threads:
            if thread.job == job:
                if not thread.started and not thread.stopped:
                    thread.start()
                    log(f"Thread '{job.name}' started.", level=5)
                    break

    def replace_thread(self, job: Job) -> None:
        log(f"Replacing thread for \"{job.name}\"", level=6)
        self.stop_thread(job)
        self.add_thread(job)
        self.start_thread(job)

    def list(self) -> list[Job]:
        return [thread.job for thread in self.threads]

    def list_pretty(self) -> list:
        pretty = []
        for thread in self.threads:
            if thread.next_execution_time is None:
                next_run = 'None'
            else:
                next_run = thread.next_execution_time

            pretty.append(f'{thread.job.name} next run at {next_run}')

        return pretty

    def clean_stopped_threads(self):
        to_remove = []
        for thread in self.threads:
            if thread.stopped:
                to_remove.append(thread)

        for thread in to_remove:
            self.threads.remove(thread)

    def __del__(self):
        try:
            self.stop()

        except ImportError:
            pass



# Generated by Django 5.0.2 on 2024-02-24 09:45

import aw.model.job
import django.db.models.deletion
import django.utils.timezone
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name="Job",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                (
                    "limit",
                    models.CharField(
                        blank=True, default=None, max_length=500, null=True
                    ),
                ),
                (
                    "verbosity",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (0, "Default"),
                            (1, "v"),
                            (2, "vv"),
                            (3, "vvv"),
                            (4, "vvvv"),
                            (5, "vvvv"),
                            (6, "vvvvvv"),
                        ],
                        default=0,
                    ),
                ),
                (
                    "comment",
                    models.CharField(
                        blank=True, default=None, max_length=150, null=True
                    ),
                ),
                (
                    "mode_diff",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=False
                    ),
                ),
                (
                    "mode_check",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=False
                    ),
                ),
                (
                    "environment_vars",
                    models.CharField(
                        blank=True, default=None, max_length=1000, null=True
                    ),
                ),
                (
                    "tags",
                    models.CharField(
                        blank=True, default=None, max_length=150, null=True
                    ),
                ),
                (
                    "tags_skip",
                    models.CharField(
                        blank=True, default=None, max_length=150, null=True
                    ),
                ),
                (
                    "cmd_args",
                    models.CharField(
                        blank=True, default=None, max_length=150, null=True
                    ),
                ),
                ("name", models.CharField(max_length=150)),
                ("inventory_file", models.CharField(max_length=300)),
                ("playbook_file", models.CharField(max_length=100)),
                (
                    "schedule",
                    models.CharField(
                        blank=True,
                        default=None,
                        max_length=50,
                        null=True,
                        validators=[aw.model.job.validate_cronjob],
                    ),
                ),
                (
                    "enabled",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=True
                    ),
                ),
                (
                    "credentials_needed",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=True
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="JobCredentialsPermissionMapping",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name="JobError",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("short", models.CharField(max_length=100)),
                ("med", models.TextField(max_length=1024, null=True)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="JobGlobalCredentials",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=100)),
                (
                    "connect_user",
                    models.CharField(
                        blank=True, default=None, max_length=100, null=True
                    ),
                ),
                ("become_user", models.CharField(default="root", max_length=100)),
                (
                    "vault_file",
                    models.CharField(
                        blank=True, default=None, max_length=300, null=True
                    ),
                ),
                (
                    "vault_id",
                    models.CharField(
                        blank=True, default=None, max_length=50, null=True
                    ),
                ),
                (
                    "_enc_vault_pass",
                    models.CharField(
                        blank=True, default=None, max_length=500, null=True
                    ),
                ),
                (
                    "_enc_become_pass",
                    models.CharField(
                        blank=True, default=None, max_length=500, null=True
                    ),
                ),
                (
                    "_enc_connect_pass",
                    models.CharField(
                        blank=True, default=None, max_length=500, null=True
                    ),
                ),
                (
                    "_enc_ssh_key",
                    models.CharField(
                        blank=True, default=None, max_length=5000, null=True
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="JobPermission",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=100)),
                (
                    "permission",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (0, "None"),
                            (5, "Read"),
                            (10, "Execute"),
                            (15, "Write"),
                            (20, "Full"),
                        ],
                        default=0,
                    ),
                ),
                (
                    "jobs_all",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=False
                    ),
                ),
                (
                    "credentials_all",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=False
                    ),
                ),
                (
                    "repositories_all",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=False
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="JobPermissionMapping",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name="JobPermissionMemberGroup",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name="JobPermissionMemberUser",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name="JobQueue",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="JobRepositoryPermissionMapping",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
            ],
        ),
        migrations.CreateModel(
            name="JobUserCredentials",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=100)),
                (
                    "connect_user",
                    models.CharField(
                        blank=True, default=None, max_length=100, null=True
                    ),
                ),
                ("become_user", models.CharField(default="root", max_length=100)),
                (
                    "vault_file",
                    models.CharField(
                        blank=True, default=None, max_length=300, null=True
                    ),
                ),
                (
                    "vault_id",
                    models.CharField(
                        blank=True, default=None, max_length=50, null=True
                    ),
                ),
                (
                    "_enc_vault_pass",
                    models.CharField(
                        blank=True, default=None, max_length=500, null=True
                    ),
                ),
                (
                    "_enc_become_pass",
                    models.CharField(
                        blank=True, default=None, max_length=500, null=True
                    ),
                ),
                (
                    "_enc_connect_pass",
                    models.CharField(
                        blank=True, default=None, max_length=500, null=True
                    ),
                ),
                (
                    "_enc_ssh_key",
                    models.CharField(
                        blank=True, default=None, max_length=5000, null=True
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="Repository",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                ("name", models.CharField(max_length=100)),
                (
                    "rtype",
                    models.PositiveSmallIntegerField(
                        choices=[(1, "Static"), (2, "Git")]
                    ),
                ),
                (
                    "time_update",
                    models.DateTimeField(blank=True, default=None, null=True),
                ),
                (
                    "status",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (0, "Waiting"),
                            (1, "Starting"),
                            (2, "Running"),
                            (3, "Failed"),
                            (4, "Finished"),
                            (5, "Stopping"),
                            (6, "Stopped"),
                        ],
                        default=0,
                    ),
                ),
                (
                    "log_stdout",
                    models.CharField(
                        blank=True, default=None, max_length=300, null=True
                    ),
                ),
                (
                    "log_stderr",
                    models.CharField(
                        blank=True, default=None, max_length=300, null=True
                    ),
                ),
                (
                    "static_path",
                    models.CharField(
                        blank=True, default=None, max_length=500, null=True
                    ),
                ),
                (
                    "git_origin",
                    models.CharField(
                        blank=True, default=None, max_length=100, null=True
                    ),
                ),
                (
                    "git_branch",
                    models.CharField(
                        blank=True, default=None, max_length=100, null=True
                    ),
                ),
                (
                    "git_isolate",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=False
                    ),
                ),
                (
                    "git_lfs",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=False
                    ),
                ),
                (
                    "git_limit_depth",
                    models.PositiveIntegerField(blank=True, default=None, null=True),
                ),
                (
                    "git_hook_pre",
                    models.CharField(
                        blank=True, default=None, max_length=1000, null=True
                    ),
                ),
                (
                    "git_hook_post",
                    models.CharField(
                        blank=True, default=None, max_length=1000, null=True
                    ),
                ),
                (
                    "git_override_initialize",
                    models.CharField(
                        blank=True, default=None, max_length=1000, null=True
                    ),
                ),
                (
                    "git_override_update",
                    models.CharField(
                        blank=True, default=None, max_length=1000, null=True
                    ),
                ),
                (
                    "git_playbook_base",
                    models.CharField(
                        blank=True, default=None, max_length=300, null=True
                    ),
                ),
            ],
        ),
        migrations.CreateModel(
            name="SystemConfig",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                (
                    "path_run",
                    models.CharField(default="/tmp/ansible-webui", max_length=500),
                ),
                ("path_play", models.CharField(default=None, max_length=500)),
                ("path_log", models.CharField(default=None, max_length=500)),
                ("timezone", models.CharField(default="CET", max_length=300)),
                ("run_timeout", models.PositiveIntegerField(default=3600)),
                ("session_timeout", models.PositiveIntegerField(default=43200)),
                (
                    "path_ansible_config",
                    models.CharField(
                        blank=True, default=None, max_length=500, null=True
                    ),
                ),
                (
                    "path_ssh_known_hosts",
                    models.CharField(
                        blank=True, default=None, max_length=500, null=True
                    ),
                ),
                (
                    "debug",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=False
                    ),
                ),
                (
                    "logo_url",
                    models.CharField(
                        blank=True, default=None, max_length=500, null=True
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="AwAPIKey",
            fields=[
                (
                    "id",
                    models.CharField(
                        editable=False,
                        max_length=150,
                        primary_key=True,
                        serialize=False,
                        unique=True,
                    ),
                ),
                ("prefix", models.CharField(editable=False, max_length=8, unique=True)),
                ("hashed_key", models.CharField(editable=False, max_length=150)),
                ("created", models.DateTimeField(auto_now_add=True, db_index=True)),
                (
                    "name",
                    models.CharField(
                        default=None,
                        help_text="A free-form name for the API key. Need not be unique. 50 characters max.",
                        max_length=50,
                    ),
                ),
                (
                    "revoked",
                    models.BooleanField(
                        blank=True,
                        default=False,
                        help_text="If the API key is revoked, clients cannot use it anymore. (This cannot be undone.)",
                    ),
                ),
                (
                    "expiry_date",
                    models.DateTimeField(
                        blank=True,
                        help_text="Once API key expires, clients cannot use it anymore.",
                        null=True,
                        verbose_name="Expires",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        editable=False,
                        on_delete=django.db.models.deletion.CASCADE,
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
            ],
            options={
                "verbose_name": "API key",
                "verbose_name_plural": "API keys",
                "ordering": ("-created",),
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="JobExecutionResult",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("time_start", models.DateTimeField(default=django.utils.timezone.now)),
                ("time_fin", models.DateTimeField(blank=True, default=None, null=True)),
                (
                    "failed",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=False
                    ),
                ),
                (
                    "error",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="jobresult_fk_error",
                        to="aw.joberror",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="JobExecution",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("updated", models.DateTimeField(auto_now=True)),
                (
                    "limit",
                    models.CharField(
                        blank=True, default=None, max_length=500, null=True
                    ),
                ),
                (
                    "verbosity",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (0, "Default"),
                            (1, "v"),
                            (2, "vv"),
                            (3, "vvv"),
                            (4, "vvvv"),
                            (5, "vvvv"),
                            (6, "vvvvvv"),
                        ],
                        default=0,
                    ),
                ),
                (
                    "comment",
                    models.CharField(
                        blank=True, default=None, max_length=150, null=True
                    ),
                ),
                (
                    "mode_diff",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=False
                    ),
                ),
                (
                    "mode_check",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=False
                    ),
                ),
                (
                    "environment_vars",
                    models.CharField(
                        blank=True, default=None, max_length=1000, null=True
                    ),
                ),
                (
                    "tags",
                    models.CharField(
                        blank=True, default=None, max_length=150, null=True
                    ),
                ),
                (
                    "tags_skip",
                    models.CharField(
                        blank=True, default=None, max_length=150, null=True
                    ),
                ),
                (
                    "cmd_args",
                    models.CharField(
                        blank=True, default=None, max_length=150, null=True
                    ),
                ),
                (
                    "status",
                    models.PositiveSmallIntegerField(
                        choices=[
                            (0, "Waiting"),
                            (1, "Starting"),
                            (2, "Running"),
                            (3, "Failed"),
                            (4, "Finished"),
                            (5, "Stopping"),
                            (6, "Stopped"),
                        ],
                        default=0,
                    ),
                ),
                (
                    "log_stdout",
                    models.CharField(
                        blank=True, default=None, max_length=300, null=True
                    ),
                ),
                (
                    "log_stderr",
                    models.CharField(
                        blank=True, default=None, max_length=300, null=True
                    ),
                ),
                (
                    "log_stdout_repo",
                    models.CharField(
                        blank=True, default=None, max_length=300, null=True
                    ),
                ),
                (
                    "log_stderr_repo",
                    models.CharField(
                        blank=True, default=None, max_length=300, null=True
                    ),
                ),
                (
                    "command",
                    models.CharField(
                        blank=True, default=None, max_length=1000, null=True
                    ),
                ),
                (
                    "job",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="jobexec_fk_job",
                        to="aw.job",
                    ),
                ),
                (
                    "user",
                    models.ForeignKey(
                        editable=False,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="jobexec_fk_user",
                        to=settings.AUTH_USER_MODEL,
                    ),
                ),
                (
                    "user_credentials",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="jobexec_fk_usercreds",
                        to="aw.job",
                    ),
                ),
                (
                    "result",
                    models.ForeignKey(
                        blank=True,
                        default=None,
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="jobexec_fk_result",
                        to="aw.jobexecutionresult",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="JobExecutionResultHost",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created", models.DateTimeField(auto_now_add=True)),
                ("hostname", models.CharField(max_length=300)),
                (
                    "unreachable",
                    models.BooleanField(
                        choices=[(True, "Yes"), (False, "No")], default=False
                    ),
                ),
                ("tasks_skipped", models.PositiveSmallIntegerField(default=0)),
                ("tasks_ok", models.PositiveSmallIntegerField(default=0)),
                ("tasks_failed", models.PositiveSmallIntegerField(default=0)),
                ("tasks_rescued", models.PositiveSmallIntegerField(default=0)),
                ("tasks_ignored", models.PositiveSmallIntegerField(default=0)),
                ("tasks_changed", models.PositiveSmallIntegerField(default=0)),
                (
                    "error",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.SET_NULL,
                        related_name="jobresulthost_fk_error",
                        to="aw.joberror",
                    ),
                ),
                (
                    "result",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="jobresulthost_fk_result",
                        to="aw.jobexecutionresult",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.AddConstraint(
            model_name="jobglobalcredentials",
            constraint=models.UniqueConstraint(fields=("name",), name="jobcreds_name"),
        ),
        migrations.AddField(
            model_name="jobexecution",
            name="credential_global",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="jobexec_fk_credglob",
                to="aw.jobglobalcredentials",
            ),
        ),
        migrations.AddField(
            model_name="jobcredentialspermissionmapping",
            name="credentials",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                to="aw.jobglobalcredentials",
            ),
        ),
        migrations.AddField(
            model_name="job",
            name="credentials_default",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="job_fk_creddflt",
                to="aw.jobglobalcredentials",
            ),
        ),
        migrations.AddField(
            model_name="jobpermission",
            name="credentials",
            field=models.ManyToManyField(
                through="aw.JobCredentialsPermissionMapping",
                to="aw.jobglobalcredentials",
            ),
        ),
        migrations.AddField(
            model_name="jobcredentialspermissionmapping",
            name="permission",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="aw.jobpermission"
            ),
        ),
        migrations.AddField(
            model_name="jobpermissionmapping",
            name="job",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="aw.job"
            ),
        ),
        migrations.AddField(
            model_name="jobpermissionmapping",
            name="permission",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="aw.jobpermission"
            ),
        ),
        migrations.AddField(
            model_name="jobpermission",
            name="jobs",
            field=models.ManyToManyField(
                through="aw.JobPermissionMapping", to="aw.job"
            ),
        ),
        migrations.AddField(
            model_name="jobpermissionmembergroup",
            name="group",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="auth.group"
            ),
        ),
        migrations.AddField(
            model_name="jobpermissionmembergroup",
            name="permission",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="aw.jobpermission"
            ),
        ),
        migrations.AddField(
            model_name="jobpermission",
            name="groups",
            field=models.ManyToManyField(
                through="aw.JobPermissionMemberGroup", to="auth.group"
            ),
        ),
        migrations.AddField(
            model_name="jobpermissionmemberuser",
            name="permission",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="aw.jobpermission"
            ),
        ),
        migrations.AddField(
            model_name="jobpermissionmemberuser",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AddField(
            model_name="jobpermission",
            name="users",
            field=models.ManyToManyField(
                through="aw.JobPermissionMemberUser", to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AddField(
            model_name="jobqueue",
            name="job",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="jobqueue_fk_job",
                to="aw.job",
            ),
        ),
        migrations.AddField(
            model_name="jobqueue",
            name="user",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="jobqueue_fk_user",
                to=settings.AUTH_USER_MODEL,
            ),
        ),
        migrations.AddField(
            model_name="jobrepositorypermissionmapping",
            name="permission",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="aw.jobpermission"
            ),
        ),
        migrations.AddField(
            model_name="jobusercredentials",
            name="user",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL
            ),
        ),
        migrations.AddField(
            model_name="jobexecution",
            name="credential_user",
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="jobexec_fk_credusr",
                to="aw.jobusercredentials",
            ),
        ),
        migrations.AddField(
            model_name="repository",
            name="git_credentials",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="repo_fk_cred",
                to="aw.jobglobalcredentials",
            ),
        ),
        migrations.AddField(
            model_name="jobrepositorypermissionmapping",
            name="repository",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE, to="aw.repository"
            ),
        ),
        migrations.AddField(
            model_name="jobpermission",
            name="repositories",
            field=models.ManyToManyField(
                through="aw.JobRepositoryPermissionMapping", to="aw.repository"
            ),
        ),
        migrations.AddField(
            model_name="job",
            name="repository",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="job_fk_repo",
                to="aw.repository",
            ),
        ),
        migrations.AddConstraint(
            model_name="jobcredentialspermissionmapping",
            constraint=models.UniqueConstraint(
                fields=("credentials", "permission"), name="jobcredpermmap_unique"
            ),
        ),
        migrations.AddConstraint(
            model_name="jobpermissionmapping",
            constraint=models.UniqueConstraint(
                fields=("job", "permission"), name="jobpermmap_unique"
            ),
        ),
        migrations.AddConstraint(
            model_name="jobpermissionmembergroup",
            constraint=models.UniqueConstraint(
                fields=("group", "permission"), name="jobpermmembergrp_unique"
            ),
        ),
        migrations.AddConstraint(
            model_name="jobpermissionmemberuser",
            constraint=models.UniqueConstraint(
                fields=("user", "permission"), name="jobpermmemberuser_unique"
            ),
        ),
        migrations.AddConstraint(
            model_name="jobusercredentials",
            constraint=models.UniqueConstraint(
                fields=("user", "name"), name="jobusercreds_user_name"
            ),
        ),
        migrations.AddConstraint(
            model_name="repository",
            constraint=models.UniqueConstraint(
                fields=("name",), name="repo_name_unique"
            ),
        ),
        migrations.AddConstraint(
            model_name="jobrepositorypermissionmapping",
            constraint=models.UniqueConstraint(
                fields=("repository", "permission"), name="jobrepopermmap_unique"
            ),
        ),
        migrations.AddConstraint(
            model_name="jobpermission",
            constraint=models.UniqueConstraint(
                fields=("name",), name="jobperm_name_unique"
            ),
        ),
        migrations.AddConstraint(
            model_name="job",
            constraint=models.UniqueConstraint(
                fields=("name",), name="job_name_unique"
            ),
        ),
    ]



from pathlib import Path
from shutil import copy
from datetime import datetime
from sys import exit as sys_exit
from secrets import choice as random_choice
from string import digits, ascii_letters
from os import listdir, remove
from time import time
from sqlite3 import connect as db_connect
from sqlite3 import OperationalError

from aw.config.main import VERSION
from aw.settings import DB_FILE
from aw.utils.subps import process
from aw.utils.debug import log, log_error, log_warn
from aw.utils.deployment import deployment_prod, is_release_version
from aw.config.hardcoded import FILE_TIME_FORMAT, GRP_MANAGER
from aw.config.environment import check_aw_env_var_true, get_aw_env_var, check_aw_env_var_is_set

DB_BACKUP_EXT = '.auto.bak'
DB_BACKUP_RETENTION_DAYS = 7

if not deployment_prod():
    DB_BACKUP_RETENTION_DAYS = 1

DB_BACKUP_RETENTION = DB_BACKUP_RETENTION_DAYS * 24 * 60 * 60


def _check_if_writable():
    try:
        test_file = DB_FILE.parent / '.awtest'
        with open(test_file, 'w', encoding='utf-8') as _file:
            _file.write('TEST')

        remove(test_file)

    except PermissionError:
        log(msg=f"Error: DB directory is not writable: '{DB_FILE.parent}'")
        sys_exit(1)


def _schema_up_to_date() -> bool:
    if not Path(DB_FILE).is_file():
        return False

    try:
        with db_connect(DB_FILE) as conn:
            return conn.execute('SELECT schema_version FROM aw_schemametadata').fetchall()[0][0] == VERSION

    except (IndexError, OperationalError):
        return False


def _get_current_schema_version() -> (str, None):
    try:
        with db_connect(DB_FILE) as conn:
            return conn.execute('SELECT schema_version FROM aw_schemametadata').fetchall()[0][0]

    except (IndexError, OperationalError):
        return None


def _update_schema_version() -> None:
    previous = _get_current_schema_version()

    with db_connect(DB_FILE) as conn:
        try:
            if previous is None:
                conn.execute(
                    "INSERT INTO aw_schemametadata (created, updated, schema_version) VALUES "
                    f"(DATETIME('now'), DATETIME('now'), '{VERSION}')",
                )

            else:
                conn.execute(
                    "UPDATE aw_schemametadata SET "
                    f"schema_version = '{VERSION}', schema_version_prev = '{previous}', "
                    "updated = DATETIME('now') WHERE id = 1"
                )

            conn.commit()

        except (IndexError, OperationalError) as err:
            log(msg=f"Error updating database schema version: '{err}'", level=3)


def install_or_migrate_db():
    log(msg=f"Using DB: {DB_FILE}", level=4)
    _check_if_writable()
    if not Path(DB_FILE).is_file():
        return install()

    return migrate()


def _manage_db(action: str, cmd: list, backup: str = None) -> dict:
    cmd2 = ['python3', 'manage.py']
    cmd2.extend(cmd)

    result = process(cmd=cmd2)

    if result['rc'] != 0:
        log_error(f'Database {action} failed!')
        log(msg=f"Error:\n{result['stderr']}", level=1)
        log(msg=f"Output:\n{result['stdout']}", level=3)

        if backup is not None:
            log_warn(
                msg=f"Trying to restore database from automatic backup: {backup} => {DB_FILE}",
                _stderr=True,
            )
            copy(src=DB_FILE, dst=f'{backup}.failed')
            copy(src=backup, dst=DB_FILE)

        else:
            sys_exit(1)

    return result


def _clean_old_db_backups():
    possible_db_backup_files = listdir(DB_FILE.parent)
    for file in possible_db_backup_files:
        if file.startswith(DB_FILE.name) and file.endswith(DB_BACKUP_EXT):
            backup_file = DB_FILE.parent / file
            backup_age = time() - backup_file.stat().st_mtime
            if backup_age > DB_BACKUP_RETENTION:
                log(msg=f"Cleaning old backup file: '{backup_file}'", level=4)
                remove(backup_file)


def install():
    log(msg=f"Initializing database {DB_FILE}..", level=3)
    _migration_needed()
    _manage_db(action='initialization', cmd=['migrate'])
    _update_schema_version()


def migrate():
    _clean_old_db_backups()

    if _migration_needed() and check_aw_env_var_true(var='db_migrate', fallback=True):
        backup = f"{DB_FILE}.{datetime.now().strftime(FILE_TIME_FORMAT)}{DB_BACKUP_EXT}"
        log(msg=f"Creating database backup: '{backup}'", level=6)
        copy(src=DB_FILE, dst=backup)

        log(msg=f"Upgrading database {DB_FILE}", level=3)
        if _manage_db(action='migration', cmd=['migrate'], backup=backup)['rc'] == 0:
            _update_schema_version()


def _migration_needed() -> bool:
    if is_release_version():
        # stable versions should only have released migrations
        return not _schema_up_to_date()

    changed = False

    for stdout in [
        _manage_db(action='schema-creation', cmd=['makemigrations'])['stdout'],
        _manage_db(action='schema-creation', cmd=['makemigrations', 'aw'])['stdout'],
    ]:
        if stdout.find('No changes detected') == -1:
            changed = True

    return changed or not _schema_up_to_date()


def create_first_superuser():
    # pylint: disable=C0415
    from aw.base import USERS
    if len(USERS.objects.filter(is_superuser=True)) == 0:
        name = get_aw_env_var('init_admin')
        pwd = get_aw_env_var('init_admin_pwd')

        if name is None:
            name = 'ansible'

        if pwd is None:
            pwd = ''.join(random_choice(ascii_letters + digits + '!.-+') for _ in range(14))

        USERS.objects.create_superuser(
            username=name,
            email=f"{name}@localhost",
            password=pwd
        )

        log_warn('No admin was found in the database!')
        if check_aw_env_var_is_set('init_admin_pwd'):
            log(msg=f"The user '{name}' was created!", level=4)

        else:
            log(msg=f"Generated user: '{name}'", level=3)
            log(msg=f"Generated pwd: '{pwd}'", level=3)
            log_warn('Make sure to change the password!')


def create_privileged_groups():
    # pylint: disable=C0415
    from django.contrib.auth.models import Group
    for grp in GRP_MANAGER.values():
        Group.objects.get_or_create(name=grp)



from os import environ
from time import tzset


def set_timezone(timezone: str):
    environ['TZ'] = timezone
    environ.setdefault('TZ', timezone)
    tzset()


def is_null(data) -> bool:
    if data is None:
        return True

    return str(data).strip() == ''


def is_set(data: str) -> bool:
    return not is_null(data)



from aw.model.job import Job
from aw.model.permission import JobPermissionMapping, JobPermissionMemberUser, JobPermissionMemberGroup, \
    CHOICE_PERMISSION_READ, JobCredentialsPermissionMapping, JobRepositoryPermissionMapping, JobPermission, \
    CHOICE_PERMISSION_WRITE, CHOICE_PERMISSION_DELETE
from aw.model.job_credential import BaseJobCredentials, JobGlobalCredentials
from aw.model.repository import Repository
from aw.base import USERS
from aw.utils.debug import log
from aw.config.hardcoded import GRP_MANAGER


def get_job_if_allowed(user: USERS, job: Job, permission_needed: int) -> (Job, None):
    if job is None:
        return None

    if not isinstance(job, Job):
        raise ValueError(f"Provided job is invalid: '{job}'")

    if has_job_permission(user=user, job=job, permission_needed=permission_needed):
        return job

    return None


def _evaluate_permission(permission: JobPermission, user: USERS, permission_needed: int) -> bool:
    # ignore permission if access-level is too low
    if permission.permission < permission_needed:
        return False

    # if one of the permissions allows the user
    if JobPermissionMemberUser.objects.filter(user=user, permission=permission).exists():
        return True

    # if one of the permissions allows a group that the user is a member of
    links = JobPermissionMemberGroup.objects.filter(permission=permission)
    if links.exists() and user.groups.filter(name__in=[
        link.group.name for link in links
    ]).exists():
        return True

    return False


def _has_permission(
        permission_links: (JobPermissionMapping, JobCredentialsPermissionMapping), permission_needed: int,
        user: USERS, permission_attr_all: str, manager: str = None,
) -> bool:
    if user.is_superuser:
        return True

    if manager is not None and \
            permission_needed in [CHOICE_PERMISSION_READ, CHOICE_PERMISSION_WRITE, CHOICE_PERMISSION_DELETE] and \
            has_manager_privileges(user=user, kind=manager):
        return True

    # 'all' permissions
    for permission in JobPermission.objects.filter(**{permission_attr_all: True}):
        if _evaluate_permission(permission=permission, user=user, permission_needed=permission_needed):
            log(
                msg=f"User '{user}' privileged ({permission_needed}) through permission {permission.name}",
                level=7,
            )
            return True

    # lined permissions
    for link in permission_links:
        if _evaluate_permission(permission=link.permission, user=user, permission_needed=permission_needed):
            log(
                msg=f"User '{user}' privileged ({permission_needed}) through permission {link.permission.name}",
                level=7,
            )
            return True

    return False


def has_job_permission(user: USERS, job: Job, permission_needed: int) -> bool:
    return _has_permission(
        permission_links=JobPermissionMapping.objects.filter(job=job),
        permission_needed=permission_needed,
        permission_attr_all='jobs_all',
        user=user,
        manager='job',
    )


def has_credentials_permission(
        user: USERS, credentials: BaseJobCredentials, permission_needed: int,
) -> bool:
    return _has_permission(
        permission_links=JobCredentialsPermissionMapping.objects.filter(credentials=credentials),
        permission_needed=permission_needed,
        permission_attr_all='credentials_all',
        user=user,
        manager='credentials',
    )


def has_repository_permission(
        user: USERS, repository: Repository, permission_needed: int,
) -> bool:
    return _has_permission(
        permission_links=JobRepositoryPermissionMapping.objects.filter(repository=repository),
        permission_needed=permission_needed,
        permission_attr_all='repositories_all',
        user=user,
        manager='repository',
    )


def get_viewable_jobs(user: USERS) -> list[Job]:
    jobs_viewable = []

    for job in Job.objects.all():
        if has_job_permission(user=user, job=job, permission_needed=CHOICE_PERMISSION_READ):
            jobs_viewable.append(job)

    return jobs_viewable


def get_viewable_credentials(user: USERS) -> list[BaseJobCredentials]:
    credentials_viewable = []

    for credentials in JobGlobalCredentials.objects.all():
        if has_credentials_permission(user=user, credentials=credentials, permission_needed=CHOICE_PERMISSION_READ):
            credentials_viewable.append(credentials)

    return credentials_viewable


def get_viewable_repositories(user: USERS) -> list[Repository]:
    repositories_viewable = []

    for repository in Repository.objects.all():
        if has_repository_permission(user=user, repository=repository, permission_needed=CHOICE_PERMISSION_READ):
            repositories_viewable.append(repository)

    return repositories_viewable


def has_manager_privileges(user: USERS, kind: str) -> bool:
    if user.is_superuser:
        return True

    return user.groups.filter(name=GRP_MANAGER[kind]).exists()



from aw.config.main import config

FORM_LABEL = {
    'jobs': {
        'manage': {
            'repository': 'Repository',
            'environment_vars': 'Environmental Variables',
            'mode_diff': 'Diff Mode',
            'mode_check': 'Check Mode (Try Run)',
            'cmd_args': 'Commandline Arguments',
            'user_credentials': 'User Credentials',
            'enabled': 'Schedule Enabled',
            'tags_skip': 'Skip Tags',
            'credentials_needed': 'Needs Credentials',
            'credentials_default': 'Default Job Credentials',
        },
        'credentials': {
            'connect_user': 'Connect User',
            'connect_pass': 'Connect Password',
            'become_user': 'Become User',
            'become_pass': 'Become Password',
            'vault_file': 'Vault Password File',
            'vault_pass': 'Vault Password',
            'vault_id': 'Vault ID',
            'ssh_key': 'SSH Private Key',
        },
        'repository': {
            'rtype': 'Repository Type',
            'static_path': 'Static Repository Path',
            'git_origin': 'Git Origin',
            'git_branch': 'Git Branch',
            'git_credentials': 'Git Credentials',
            'git_limit_depth': 'Git Limit Depth',
            'git_lfs': 'Git LFS',
            'git_playbook_base': 'Git Playbook Base-Directory',
            'git_isolate': 'Git Isolate Directory',
            'git_hook_pre': 'Git Pre-Hook',
            'git_hook_post': 'Git Post-Hook',
            'git_override_initialize': 'Git Override Initialize-Command',
            'git_override_update': 'Git Override Update-Command',
        },
    },
    'settings': {
        'permissions': {
            'jobs_all': 'All jobs',
            'credentials_all': 'All credentials',
            'repositories_all': 'All repositories',
        },
    },
    'system': {
        'config': {
            'path_run': 'Runtime directory',
            'path_play': 'Playbook base-directory',
            'path_log': 'Directory for execution-logs',
            'run_timeout': 'Timeout for playbook execution',
            'session_timeout': 'Timeout for WebUI login-sessions',
            'path_ansible_config': 'Ansible Config-File',
            'path_ssh_known_hosts': 'SSH Known-Hosts File',
            'debug': 'Debug Mode',
            # env-vars
            'timezone': 'Timezone',
            'db': 'Database',
            'db_migrate': 'Database auto-upgrade',
            'serve_static': 'Serving static files',
            'deployment': 'Deployment',
            'version': 'AnsibleGuy WebUI Version',
            'logo_url': 'URL to a Logo to use',
            'ara_server': 'ARA Server URL',
            'global_environment_vars': 'Global Environmental Variables',
            'auth_mode': 'Authentication Mode',
            'saml_config_file': 'SAML Config File',
        }
    }
}

FORM_HELP = {
    'jobs': {
        'manage': {
            'playbook_file': f"Playbook to execute. Search path: '{config['path_play']}'",
            # todo: change search-path with repository
            'inventory_file': 'One or multiple inventory files/directories to include for the execution. '
                              'Comma-separated list. For details see: '
                              '<a href="https://docs.ansible.com/ansible/latest/inventory_guide/'
                              'intro_inventory.html">Ansible Docs - Inventory</a>',
            'repository': 'Used to define the static or dynamic source of your playbook directory structure. '
                          f"Default is '{config['path_play']}'",
            'limit': 'Ansible inventory hosts or groups to limit the execution to.'
                     'For details see: '
                     '<a href="https://docs.ansible.com/ansible/latest/inventory_guide/intro_patterns.html">'
                     'Ansible Docs - Limit</a>',
            'schedule': 'Schedule for running the job automatically. For format see: '
                        '<a href="https://crontab.guru/">crontab.guru</a>',
            'environment_vars': 'Environmental variables to be passed to the Ansible execution. '
                                'Comma-separated list of key-value pairs. (VAR1=TEST1,VAR2=0)',
            'cmd_args': "Additional commandline arguments to pass to 'ansible-playbook'. "
                        "Can be used to pass extra-vars",
            'tags': 'For details see: '
                    '<a href="https://docs.ansible.com/ansible/latest/playbook_guide/playbooks_tags.html">'
                    'Ansible Docs - Tags</a>',
            'mode_check': 'For details see: '
                          '<a href="https://docs.ansible.com/ansible/2.8/user_guide/playbooks_checkmode.html">'
                          'Ansible Docs - Check Mode</a>',
            'credentials_needed': 'If the job requires credentials to be specified '
                                  '(either as default or at execution-time; '
                                  'fallback are the user-credentials of the executing user)',
            'credentials_default': 'Specify job-level default credentials to use',
            'enabled': 'En- or disable the schedule. Can be ignored if no schedule was set',
        },
        'credentials': {
            'vault_file': 'Path to the file containing your vault-password',
            'vault_id': 'For details see: '
                        '<a href="https://docs.ansible.com/ansible/latest/vault_guide/'
                        'vault_managing_passwords.html">'
                        'Ansible Docs - Managing Passwords</a>',
            'ssh_key': 'Provide an unencrypted SSH private key',
        },
        'repository': {
            'static_path': 'Path to the local static repository/playbook-base-directory',
            'git_origin': "Full URL to the remote repository. "
                          "Per example: '<a href=\"https://github.com/ansibleguy/webui.git\">"
                          "https://github.com/ansibleguy/webui.git'</a>'",
            'git_credentials': "Credentials for connecting to the origin. "
                               "'Connect User', 'Connect Password' and 'SSH Private Key' are used",
            'git_playbook_base': 'Relative path to the Playbook base-directory relative from the repository root',
            'git_lfs': 'En- or disable checkout of Git-LFS files',
            'git_isolate': 'En- or disable if one clone of the Git-repository should be used for all jobs. '
                           'If enabled - the repository will be cloned/fetched on every job execution. '
                           'This will have a negative impact on performance',
            'git_hook_pre': 'Commands to execute before initializing/updating the repository. '
                            'Comma-separated list of shell-commands',
            'git_hook_post': 'Commands to execute after initializing/updating the repository. '
                             'Comma-separated list of shell-commands',
            'git_override_initialize': 'Advanced usage! Completely override the command used to initialize '
                                       '(clone) the repository',
            'git_override_update': 'Advanced usage! Completely override the command used to update '
                                   '(pull) the repository',
        },
    },
    'settings': {
        'permissions': {
            'jobs_all': 'Match permission to all existing jobs (present and future)',
            'credentials_all': 'Match permission to all existing credentials (present and future)',
            'repositories_all': 'Match permission to all existing repositories (present and future)',
        },
    },
    'system': {
        'config': {
            'path_run': 'Base directory for <a href="https://ansible.readthedocs.io/projects/runner/en/latest/intro/">'
                        'Ansible-Runner runtime files</a>',
            'path_play': 'Path to the <a href="https://docs.ansible.com/ansible/2.8/user_guide/'
                         'playbooks_best_practices.html#directory-layout">Ansible base/playbook directory</a>',
            'path_log': 'Define the path where full job-logs are saved',
            'path_ansible_config': 'Path to a <a href="https://docs.ansible.com/ansible/latest/installation_guide'
                                   '/intro_configuration.html#configuration-file">Ansible config-file</a> to use',
            'path_ssh_known_hosts': 'Path to a <a href="https://en.wikibooks.org/wiki/OpenSSH/'
                                    'Client_Configuration_Files#~/.ssh/known_hosts">SSH known_hosts file</a> to use',
            'debug': 'Enable Debug-mode. Do not enable permanent on production systems! '
                     'It can possibly open attack vectors. '
                     'You might need to restart the application to apply this setting',
            'logo_url': 'Default: <a href="/static/img/logo.svg">img/logo.svg</a>; '
                        'Per example: '
                        '<a href="https://raw.githubusercontent.com/ansible/logos/main/vscode-ansible-logo'
                        '/vscode-ansible.svg">'
                        'https://raw.githubusercontent.com/ansible/logos/main/vscode-ansible-logo/vscode-ansible.svg'
                        '</a>',
            'ara_server': 'Provide the URL to your ARA server. Can be used to gather job statistics. See: '
                          '<a href="https://webui.ansibleguy.net/en/latest/usage/integrations.html">'
                          'Documentation - Integrations</a>',
            'global_environment_vars': 'Set environmental variables that will be added to every job execution. '
                                       'Comma-separated list of key-value pairs. (VAR1=TEST1,VAR2=0)',
        }
    }
}



#!/usr/bin/env python3

if __name__ == '__main__':
    # pylint: disable=E0401
    from sys import argv as sys_argv
    from sys import exit as sys_exit
    from sys import path as sys_path
    from os import path as os_path

    try:
        from main import main

    except ModuleNotFoundError:
        sys_path.append(os_path.dirname(os_path.abspath(__file__)))
        from main import main

    if len(sys_argv) > 1:
        if sys_argv[1] in ['--version', '-v']:
            from cli_init import init_cli
            from cli import _print_version
            init_cli()
            _print_version()
            sys_exit(0)

    from aw.config.main import VERSION
    print(f'AnsibleGuy-WebUI Version {VERSION}')
    main()



from pathlib import Path

from django.core.exceptions import ObjectDoesNotExist

from aw.model.job import Job, JobExecution
from aw.model.job_credential import BaseJobCredentials, JobUserCredentials
from aw.utils.permission import has_credentials_permission, CHOICE_PERMISSION_READ
from aw.base import USERS
from aw.utils.debug import log  # log_warn
from aw.utils.util import is_set, is_null
from aw.execute.util import config_error, write_file_0600


def get_pwd_file(path_run: (str, Path), attr: str) -> str:
    return f"{path_run}/.aw_{attr}"


def get_pwd_file_arg(credentials: BaseJobCredentials, attr: str, path_run: (Path, str)) -> (str, None):
    if is_null(getattr(credentials, attr)):
        return None

    return f"--{BaseJobCredentials.SECRET_ATTRS_ARGS[attr]} {get_pwd_file(path_run=path_run, attr=attr)}"


def write_pwd_file(credentials: BaseJobCredentials, attr: str, path_run: (Path, str)):
    if credentials is None or is_null(getattr(credentials, attr)):
        return None

    return write_file_0600(
        file=get_pwd_file(path_run=path_run, attr=attr),
        content=getattr(credentials, attr),
    )


def _scheduled_or_has_credentials_access(user: USERS, credentials: BaseJobCredentials) -> bool:
    if user is None:
        # scheduled execution; permission has been checked at creation-time
        # todo: add 'owner' to jobs so we can re-check the permissions of scheduled jobs (?)
        return True

    permitted = has_credentials_permission(
        user=user,
        credentials=credentials,
        permission_needed=CHOICE_PERMISSION_READ,
    )
    if not permitted:
        log(
            msg=f"User '{user.username}' has no permission to use credentials {credentials.name}",
            level=7,
        )

    return permitted


def get_credentials_to_use(job: Job, execution: JobExecution) -> (BaseJobCredentials, None):
    credentials = None

    # todo: write warn log if user tried to execute job using non-permitted credentials (if execution.cred*)
    if execution.user is not None and is_set(execution.credential_user) and \
            execution.credential_user.user.id == execution.user.id:
        credentials = execution.credential_user

    elif is_set(execution.credential_global) and _scheduled_or_has_credentials_access(
        user=execution.user, credentials=execution.credential_global,
    ):
        credentials = execution.credential_global

    elif is_set(job.credentials_default) and _scheduled_or_has_credentials_access(
        user=execution.user, credentials=job.credentials_default,
    ):
        credentials = job.credentials_default

    elif job.credentials_needed and is_set(execution.user):
        # try to get default user-credentials as a last-resort if the job needs some credentials
        try:
            credentials = JobUserCredentials.objects.filter(user=execution.user).first()

        except ObjectDoesNotExist:
            pass

    if job.credentials_needed and credentials is None:
        config_error(
            'The job is set to require credentials - but none were provided or readable! '
            'Make sure you have privileges for the configured credentials or create user-specific ones.'
        )

    return credentials


def commandline_arguments_credentials(credentials: BaseJobCredentials, path_run: Path) -> list:
    cmd_arguments = []

    for attr, flag in BaseJobCredentials.PUBLIC_ATTRS_ARGS.items():
        if is_set(getattr(credentials, attr)):
            cmd_arguments.append(f'{flag} {getattr(credentials, attr)}')

    for attr in BaseJobCredentials.SECRET_ATTRS:
        pwd_arg = get_pwd_file_arg(credentials, attr=attr, path_run=path_run)
        if pwd_arg is not None:
            cmd_arguments.append(pwd_arg)

    return cmd_arguments



from hashlib import md5

from django.core.exceptions import ObjectDoesNotExist
from rest_framework.views import APIView
from rest_framework import serializers
from rest_framework.response import Response
from drf_spectacular.utils import extend_schema, OpenApiResponse

from aw.utils.util import datetime_w_tz
from aw.config.hardcoded import KEY_TIME_FORMAT
from aw.model.api import AwAPIKey
from aw.api_endpoints.base import API_PERMISSION, get_api_user, BaseResponse, GenericResponse


class KeyReadResponse(BaseResponse):
    token = serializers.CharField()
    id = serializers.CharField()


class KeyWriteResponse(BaseResponse):
    token = serializers.CharField()
    secret = serializers.CharField()


class APIKey(APIView):
    http_method_names = ['post', 'get']
    serializer_class = KeyReadResponse
    permission_classes = API_PERMISSION

    @staticmethod
    @extend_schema(
        request=None,
        responses={200: KeyReadResponse},
        summary='Return a list of all existing API keys of the current user.',
    )
    def get(request):
        tokens = []
        for key in AwAPIKey.objects.filter(user=get_api_user(request)):
            tokens.append({'token': key.name, 'id': md5(key.name.encode('utf-8')).hexdigest()})

        return Response(tokens)

    @extend_schema(
        request=None,
        responses={200: OpenApiResponse(KeyWriteResponse, description='Returns generated API token & key')},
        summary='Create a new API key.',
    )
    def post(self, request):
        self.serializer_class = KeyWriteResponse
        user = get_api_user(request)
        token = f'{user}-{datetime_w_tz().strftime(KEY_TIME_FORMAT)}'
        _, key = AwAPIKey.objects.create_key(name=token, user=user)
        return Response({'token': token, 'key': key})


class APIKeyItem(APIView):
    http_method_names = ['delete']
    serializer_class = GenericResponse
    permission_classes = API_PERMISSION

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(response=GenericResponse, description='API key deleted'),
            404: OpenApiResponse(response=GenericResponse, description='API key does not exist'),
        },
        summary='Delete one of the existing API keys of the current user.',
    )
    def delete(self, request, token: str):
        try:
            result = AwAPIKey.objects.get(user=get_api_user(request), name=token)

            if result is not None:
                result.delete()
                return Response(data={'msg': 'API key deleted'}, status=200)

        except ObjectDoesNotExist:
            pass

        return Response(data={'msg': 'API key not found'}, status=404)



from os import environ
from functools import cache
from pathlib import Path

from aw.config.defaults import CONFIG_DEFAULTS

AW_ENV_VARS = {
    'port': ['AW_PORT'],
    'address': ['AW_LISTEN', 'AW_LISTEN_ADDRESS'],
    'timezone': ['AW_TIMEZONE'],
    'secret': ['AW_SECRET'],
    'path_run': ['AW_PATH_RUN'],
    'path_play': ['AW_PATH_PLAY', 'ANSIBLE_PLAYBOOK_DIR'],
    'version': ['AW_VERSION'],
    'deployment': ['AW_ENV'],
    'serve_static': ['AW_SERVE_STATIC'],
    'init_admin': ['AW_ADMIN'],
    'init_admin_pwd': ['AW_ADMIN_PWD'],
    'db': ['AW_DB'],
    'db_migrate': ['AW_DB_MIGRATE'],
    'run_timeout': ['AW_RUN_TIMEOUT'],
    'path_ansible_config': ['ANSIBLE_CONFIG'],
    'path_log': ['AW_PATH_LOG'],
    'session_timeout': ['AW_SESSION_TIMEOUT'],
    'path_ssh_known_hosts': ['AW_SSH_KNOWN_HOSTS'],
    'ssl_file_crt': ['AW_SSL_CERT'],
    'ssl_file_key': ['AW_SSL_KEY'],
    'debug': ['AW_DEBUG'],
    'auth_mode': ['AW_AUTH'],
    'saml_config_file': ['AW_SAML_CONFIG'],
}
AW_ENV_VARS_SECRET = ['secret', 'init_admin', 'init_admin_pwd']

AW_ENV_VARS_REV = {}
for key_config, keys_env in AW_ENV_VARS.items():
    for key_env in keys_env:
        AW_ENV_VARS_REV[key_env] = key_config


def get_aw_env_var(var: str) -> (str, None):
    if var in AW_ENV_VARS:
        for key in AW_ENV_VARS[var]:
            if key in environ:
                return environ[key]

    return None


@cache
def get_aw_env_var_or_default(var: str) -> (str, list, None):
    val = get_aw_env_var(var)
    if val is None:
        val = CONFIG_DEFAULTS.get(var, None)

    return val


def check_aw_env_var_is_set(var: str) -> bool:
    return get_aw_env_var(var) is not None


# only use on edge-cases; as.config.main.Config.is_true is preferred
def check_aw_env_var_true(var: str, fallback: bool = False) -> bool:
    val = get_aw_env_var_or_default(var)
    if val is None:
        return fallback

    return str(val).lower() in ['1', 'true', 'y', 'yes']


def auth_mode_saml() -> bool:
    saml_config_file = get_aw_env_var_or_default('saml_config_file')
    return get_aw_env_var_or_default('auth_mode').lower() == 'saml' and \
        saml_config_file is not None and Path(saml_config_file).is_file()



from os import environ, getcwd
from pathlib import Path
from secrets import choice as random_choice
from string import digits, ascii_letters, punctuation
from datetime import datetime


def inside_docker() -> bool:
    return 'AW_DOCKER' in environ and environ['AW_DOCKER'] == '1'


def _get_existing_ansible_config_file() -> (str, None):
    # https://docs.ansible.com/ansible/latest/reference_appendices/config.html#the-configuration-file

    for file in [
        getcwd() + '/ansible.cfg',
        environ['HOME'] + '/ansible.cfg',
        environ['HOME'] + '/.ansible.cfg',
        '/etc/ansible/ansible.cfg',
    ]:
        if Path(file).is_file():
            return file

    return None


def _get_defaults_docker(var: str) -> any:
    if not inside_docker():
        return None

    return {
        'path_ssh_known_hosts': f'{getcwd()}/known_hosts',
    }[var]


# need to be referenced multiple times without import dependencies
CONFIG_DEFAULTS = {
    'port': 8000,
    'address': '127.0.0.1',
    'run_timeout': 3600,
    'path_run': '/tmp/ansible-webui',
    'path_play': getcwd(),
    'path_log': f"{environ['HOME']}/.local/share/ansible-webui",
    'db': f"{environ['HOME']}/.config/ansible-webui",
    'timezone': datetime.now().astimezone().tzname(),
    'secret': ''.join(random_choice(ascii_letters + digits + punctuation) for _ in range(50)),
    'session_timeout': 12 * 60 * 60,  # 12h
    'path_ansible_config': _get_existing_ansible_config_file(),
    'path_ssh_known_hosts': _get_defaults_docker('path_ssh_known_hosts'),
    'debug': False,
    'logo_url': 'img/logo.svg',
    'ssl_file_key': None,
    'ssl_file_crt': None,
    'auth_mode': 'local',
    'saml_config_file': None,
    'jwt_algo': 'HS256',
    'jwt_secret': ''.join(random_choice(ascii_letters + digits + punctuation) for _ in range(30)),
}



from os import environ
from os import path as os_path
from sys import path as sys_path
from importlib.metadata import version, PackageNotFoundError

from django import setup as django_setup


def init_cli():
    environ.setdefault('AW_INIT', '1')
    # workaround for CI
    if 'AW_VERSION' not in environ:
        try:
            environ['AW_VERSION'] = version('ansible-webui')

        except PackageNotFoundError:
            environ['AW_VERSION'] = '0.0.0'

    # pylint: disable=E0401,C0415
    try:
        from aw.config.main import init_config

    except ModuleNotFoundError:
        sys_path.append(os_path.dirname(os_path.abspath(__file__)))
        from aw.config.main import init_config

    init_config()
    from aw.utils.debug import warn_if_development
    warn_if_development()

    django_setup()



from aw.config.environment import get_aw_env_var
from aw.config.defaults import inside_docker
from aw.config.main import VERSION


def deployment_dev() -> bool:
    return get_aw_env_var('deployment') == 'dev'


def deployment_staging() -> bool:
    return get_aw_env_var('deployment') == 'staging'


def deployment_prod() -> bool:
    return not deployment_dev() and not deployment_staging()


def deployment_docker() -> bool:
    return inside_docker()


def is_release_version() -> bool:
    return VERSION not in ['dev', 'staging', 'latest', '0.0.0'] and \
            VERSION.find('dev') == -1 and VERSION.find('staging') == -1 and \
            VERSION.find('latest') == -1



from platform import uname
from os import environ, getpid
from sys import exit as sys_exit

from django import setup as django_setup

from aw.config.main import init_config

init_config()

# pylint: disable=C0413,C0415
from aw.config.main import config
from aw.utils.debug import log_warn, log_error
from aw.config.hardcoded import MIN_SECRET_LEN

from db import install_or_migrate_db


def _check_for_bad_config():
    if 'AW_SECRET' not in environ:
        log_warn(
            "The environmental variable 'AW_SECRET' was not supplied! "
            "Job-secrets like passwords might not be loadable after restart."
        )

    secret_len = len(config['secret'])
    if secret_len < MIN_SECRET_LEN:
        log_error(f"The provided secret key is too short! ({secret_len}<{MIN_SECRET_LEN} characters)")
        sys_exit(1)


def main():
    if uname().system.lower() != 'linux':
        raise SystemError('Currently only linux systems are supported!')

    _check_for_bad_config()

    environ['AW_INIT'] = '1'
    environ['MAINPID'] = str(getpid())
    install_or_migrate_db()

    django_setup()
    environ['AW_INIT'] = '0'

    from db import create_first_superuser, create_privileged_groups
    from webserver import init_webserver
    from aw.execute.scheduler import init_scheduler
    from aw.utils.debug import log
    from aw.settings import AUTH_MODE

    log(msg=f"Using Auth-Mode: {AUTH_MODE}", level=4)

    create_first_superuser()
    create_privileged_groups()
    init_scheduler()
    init_webserver()



from django import template

from django.forms import BoundField, MultipleChoiceField
from django.forms.widgets import Select
from django.core.validators import RegexValidator

from aw.model.job_credential import BaseJobCredentials
from aw.utils.util import is_set
from aw.views.validation import AW_VALIDATIONS

register = template.Library()


@register.filter
def get_form_field_attributes(bf: BoundField) -> str:
    attr_str = ''
    for key, value in bf.field.widget_attrs(bf.field.widget).items():
        attr_str += f' {key}="{value}"'

    return attr_str


@register.filter
def get_form_field_validators(bf: BoundField) -> str:
    attr_str = ''

    for validator in bf.field.validators:
        if isinstance(validator, RegexValidator):
            # pylint: disable=W0212
            # was not able to get the raw regex from regex attribute ('_lazy_re_compile')
            regex = validator._constructor_args[1]['regex']
            attr_str += f' pattern="{regex}"'
            if validator.message is not None:
                attr_str += f' title="{validator.message}"'

    return attr_str


@register.filter
def form_field_is_dropdown(bf: BoundField) -> bool:
    return isinstance(bf.field.widget, Select)


def get_form_required(bf: BoundField) -> str:
    return ' required' if bf.field.required else ''


def get_form_field_value(bf: BoundField, existing: dict) -> (str, None):
    # SECRET_ATTRS are not exposed here
    if bf.name not in existing and bf.name not in BaseJobCredentials.SECRET_ATTRS:
        return None

    if bf.name in BaseJobCredentials.SECRET_ATTRS:
        enc_field = '_enc_' + bf.name
        if enc_field in existing and not is_set(existing[enc_field]):
            return None

        if enc_field not in existing:
            value = None

        else:
            value = BaseJobCredentials.SECRET_HIDDEN

    else:
        if existing[bf.name] is None:
            return None

        value = str(existing[bf.name])

    if value is None or value == 'None':
        return None

    return str(value)


@register.filter
def get_form_field_select(bf: BoundField, existing: dict) -> str:
    if not hasattr(bf.field, '_choices'):
        raise AttributeError(f"Field '{bf.name}' is of an invalid type: {type(bf.field)} - {bf.field.__dict__}")

    options_str = f'<select class="form-control" id="{bf.id_for_label}" name="{bf.name}"'

    if isinstance(bf.field, MultipleChoiceField):
        options_str += ' multiple>'
        # pylint: disable=W0212
        for option in bf.field._choices:
            is_selected = ''
            if bf.name in existing and option[0] in existing[bf.name]:
                is_selected = 'selected'

            options_str += f'<option type="checkbox" value="{option[0]}" {is_selected}>{option[1]}</option>'

    else:
        options_str += '>'
        selected = None
        if bf.name in existing:
            selected = str(existing[bf.name])
        elif f'{bf.name}_id' in existing:
            selected = str(existing[f'{bf.name}_id'])
        elif bf.field.initial is not None:
            selected = bf.field.initial

        if bf.field.required:
            if selected is None:
                options_str += '<option disabled selected value>Choose an option</option>'

        else:
            if selected is None:
                options_str += '<option selected value>None</option>'
            else:
                options_str += '<option value>None</option>'

        # pylint: disable=W0212
        for option in bf.field._choices:
            is_selected = 'selected' if str(option[0]) == str(selected) else ''
            options_str += f'<option value="{option[0]}" {is_selected}>{option[1]}</option>'

    options_str += '</select>'
    return options_str


@register.filter
def get_form_field_input(bf: BoundField, existing: dict) -> str:
    field_classes = 'form-control'
    field_attrs = f'id="{bf.id_for_label}" name="{bf.name}"'
    search_choices = ''

    field_value = ''
    value = get_form_field_value(bf, existing)
    if value is not None:
        field_value = f'value="{value}"'
    elif bf.field.initial is not None:
        field_value = f'value="{bf.field.initial}"'

    if bf.name.find('_pass') != -1 or bf.name.find('_key') != -1:
        field_attrs += ' type="password"'

    elif bf.name in AW_VALIDATIONS['file_system_browse']:
        field_classes += ' aw-fs-browse'
        field_attrs += f' type="text" aw-fs-choices="aw-fs-choices-{bf.name}"'
        search_choices = f'<ul class="aw-fs-choices" id="aw-fs-choices-{bf.name}" hidden="hidden"></ul>'

    elif bf.name in AW_VALIDATIONS['file_system_exists']:
        field_classes += ' aw-fs-exists'
        field_attrs += ' type="text"'

    return (f'<input class="{field_classes}" {field_attrs} '
            f'{field_value} {get_form_required(bf)}'
            f'{get_form_field_attributes(bf)} {get_form_field_validators(bf)}>'
            f'{search_choices}')



from os import environ
from pathlib import Path
from getpass import getuser

from django.urls import path
from django.shortcuts import HttpResponse
from django.shortcuts import render
from django.contrib.auth.decorators import login_required
from ansible_runner.interface import get_ansible_config

from aw.utils.http import ui_endpoint_wrapper
from aw.utils.subps import process_cache
from aw.utils.version import get_system_versions, parsed_ansible_version, parsed_python_modules
from aw.views.forms.system import system_config


def _parsed_ansible_collections() -> dict:
    result = process_cache('ansible-galaxy collection list')
    if result['rc'] != 0:
        return {}

    collections = {}
    col_counter = {}
    collection_path = ''
    for line in result['stdout'].split('\n'):
        if line.startswith('#'):
            collection_path = line[1:]
            continue

        if line.find('.') == -1:
            continue

        name, version = line.split(' ', 1)
        name, version = name.strip(), version.strip()
        url = f"https://galaxy.ansible.com/ui/repo/published/{name.replace('.', '/')}"

        if name in collections:
            if name in col_counter:
                col_counter[name] += 1
            else:
                col_counter[name] = 2

            name = f'{name} ({col_counter[name]})'

        collections[name] = {'version': version, 'path': collection_path, 'url': url}

    return dict(sorted(collections.items()))


def _parsed_ansible_config() -> dict:
    environ['ANSIBLE_FORCE_COLOR'] = '0'
    ansible_config_raw = get_ansible_config(action='dump', quiet=True)[0].split('\n')
    environ['ANSIBLE_FORCE_COLOR'] = '1'
    ansible_config = {}

    for line in ansible_config_raw:
        try:
            setting_comment, value = line.split('=', 1)

        except ValueError:
            continue

        setting_comment, value = setting_comment.strip(), value.strip()
        try:
            setting, comment = setting_comment.rsplit('(', 1)
            comment = comment.replace(')', '')

        except ValueError:
            setting, comment = setting_comment, '-'

        url = ("https://docs.ansible.com/ansible/latest/reference_appendices/config.html#"
               f"{setting.lower().replace('_', '-')}")

        ansible_config[setting] = {'value': value, 'comment': comment, 'url': url}

    return dict(sorted(ansible_config.items()))


def _parsed_aws_versions() -> dict:
    versions = {}
    if Path('/usr/bin/session-manager-plugin').is_file():
        versions['session-manager-plugin'] = process_cache('/usr/bin/session-manager-plugin --version')['stdout']

    if Path('/usr/bin/aws').is_file():
        versions['aws-cli'] = process_cache('/usr/bin/aws --version')['stdout']

    return versions


def _parsed_ara_version(python_modules: dict) -> (str, None):
    if 'ara' not in python_modules:
        return None

    return python_modules['ara']['version']


def _parsed_ansible_playbook() -> str:
    ap = process_cache('which ansible-playbook')
    if ap['rc'] != 0:
        return 'Not Found'

    return ap['stdout']


@login_required
@ui_endpoint_wrapper
def system_environment(request) -> HttpResponse:
    # todo: allow to check for updates (pypi, ansible-galaxy & github api)
    python_modules = parsed_python_modules()
    ansible_version = parsed_ansible_version(python_modules)
    env_system = get_system_versions(python_modules=python_modules, ansible_version=ansible_version)

    return render(
        request, status=200, template_name='system/environment.html',
        context={
            **env_system,
            'env_user': getuser(),
            'env_system': env_system,
            'env_aws': _parsed_aws_versions(),
            'env_ara': _parsed_ara_version(python_modules),
            'env_python_modules': python_modules,
            'env_ansible_config': _parsed_ansible_config(),
            'env_ansible_playbook': _parsed_ansible_playbook(),
            # 'env_ansible_roles': get_role_list(),
            'env_ansible_collections': _parsed_ansible_collections(),
        },
    )


urlpatterns_system = [
    path('ui/system/environment', system_environment),
    path('ui/system/config', system_config),
]



#!/usr/bin/env python3

from sys import exit as sys_exit
from sys import argv as sys_argv
from sys import path as sys_path
from os import path as os_path
from os import geteuid, stat, listdir
from argparse import ArgumentParser
from json import dumps as json_dumps
from pathlib import Path

# pylint: disable=C0415


def _api_key(username: str):
    # python3 -m ansibleguy-webui.cli api-key.create
    from aw.base import USERS
    from aw.model.api import AwAPIKey
    from aw.utils.util import datetime_w_tz
    from aw.config.hardcoded import KEY_TIME_FORMAT

    user = USERS.objects.get(username=username)
    token = f'{user}-{datetime_w_tz().strftime(KEY_TIME_FORMAT)}'
    _, key = AwAPIKey.objects.create_key(name=token, user=user)
    print(f'API Key created:\nToken={token}\nKey={key}')
    sys_exit(0)


def _print_version():
    # python3 -m ansibleguy-webui.cli --version
    from aw.utils.version import get_version, get_system_versions

    print(f'Version: {get_version()}\n{json_dumps(get_system_versions(), indent=4)}')
    sys_exit(0)


def _list_migrations(module_base: str):
    print('Migrations')
    migrations = listdir(Path(module_base) / 'aw' / 'migrations')
    migrations.sort()
    for mig in migrations:
        if mig.find('_v') != -1:
            version = mig.split('_', 1)[1][1:-3].replace('_', '.')
            print(f"  Version: {version}: {mig}")

    sys_exit(0)


def main():
    this_file = os_path.abspath(__file__)
    file_owner_uid = stat(this_file).st_uid
    if geteuid() not in [0, file_owner_uid]:
        print('ERROR: Only root and the code-owner are permitted to run this script!')
        sys_exit(1)

    module_base = os_path.dirname(this_file)
    # pylint: disable=E0401,C0415
    try:
        from cli_init import init_cli

    except ModuleNotFoundError:
        sys_path.append(module_base)
        from cli_init import init_cli

    init_cli()

    if len(sys_argv) > 1:
        if sys_argv[1] in ['-v', '--version']:
            _print_version()

    parser = ArgumentParser()
    parser.add_argument(
        '-a', '--action', type=str, required=True,
        choices=['api-key.create', 'migrations.list'],
    )
    parser.add_argument('-p', '--parameter', type=str, required=False)
    args = parser.parse_args()

    if args.action == 'api-key.create' and args.parameter is not None:
        _api_key(username=args.parameter)

    if args.action == 'migrations.list':
        _list_migrations(module_base)


if __name__ == '__main__':
    main()



from django.contrib.auth.decorators import login_required
from django.shortcuts import redirect, render
from django.contrib.auth.views import logout_then_login
from django.shortcuts import HttpResponse
from django.urls import path, re_path

from aw.config.hardcoded import LOGIN_PATH
from aw.settings import LOGIN_REDIRECT_URL
from aw.utils.http import ui_endpoint_wrapper
from aw.views.settings import urlpatterns_settings
from aw.views.job import urlpatterns_jobs
from aw.views.system import urlpatterns_system


def _local_iframe(_path: str, title: str) -> str:
    return f'<iframe src="{_path}" title="{title}"></iframe>'


@login_required
@ui_endpoint_wrapper
def admin(request) -> HttpResponse:
    return render(request, status=200, template_name='fallback.html', context={
        'content': _local_iframe('/_admin/', title='Admin')
    })


@login_required
@ui_endpoint_wrapper
def api_docs(request) -> HttpResponse:
    return render(request, status=200, template_name='fallback.html', context={
        'content': _local_iframe('/api/_docs', title='API Docs')
    })


@login_required
@ui_endpoint_wrapper
def not_implemented(request) -> HttpResponse:
    return render(request, status=404, template_name='fallback.html', context={'content': 'Not yet implemented'})


@ui_endpoint_wrapper
def catchall(request) -> HttpResponse:
    if request.user.is_authenticated:
        return redirect(LOGIN_REDIRECT_URL)

    return redirect(LOGIN_PATH)  # will be done by endpoint_wrapper


@login_required
@ui_endpoint_wrapper
def logout(request) -> HttpResponse:
    return logout_then_login(request)


urlpatterns_ui = [
    path('ui/system/admin/', admin),
    path('ui/system/api_docs', api_docs),
]
urlpatterns_ui += urlpatterns_jobs
urlpatterns_ui += urlpatterns_settings
urlpatterns_ui += urlpatterns_system
urlpatterns_ui += [
    path('ui/', not_implemented),
    re_path(r'^ui/*', not_implemented),
]



from functools import cache

from django import template

from aw.config.main import config
from aw.settings import STATIC_URL, AUTH_MODE
from aw.config.navigation import NAVIGATION
from aw.utils.version import get_version as get_version_util

register = template.Library()


@register.simple_tag
def get_version() -> str:
    return get_version_util()


@register.simple_tag
def get_logo() -> str:
    url = config['logo_url']
    if not url.startswith('http'):
        return f"{STATIC_URL}{url}"

    return config['logo_url']


@register.simple_tag
def set_var(val):
    return val


@register.filter
def auth_sso(_) -> bool:
    return AUTH_MODE == 'saml'


@register.filter
def get_full_uri(request):
    return request.build_absolute_uri()


@cache
@register.filter
def get_nav(key: str) -> dict:
    # serves navigation config to template
    return NAVIGATION[key]


@register.filter
def get_type(value):
    return str(type(value)).replace("<class '", '').replace("'>", '')


@register.filter
def get_value(data: dict, key: (str, int)):
    if hasattr(data, 'get'):
        return data.get(key, None)

    if hasattr(data, key):
        return getattr(data, key)

    return None


@register.filter
def get_fallback(data, fallback):
    return data if data not in [None, ''] else fallback


@register.filter
def exists(data: (dict, list, str, bool)) -> bool:
    if data is None:
        return False

    if isinstance(data, bool):
        return data

    if isinstance(data, (list, dict)):
        return len(data) > 0

    if isinstance(data, str):
        return data.strip() != ''

    return False


@register.filter
def get_choice(choices: list[tuple[int, any]], idx: int):
    return choices[idx][1]


@register.filter
def to_dict(data):
    return data.__dict__


@register.filter
def ignore_none(data):
    if data is None:
        return ''

    return data


@register.filter
def capitalize(data: str) -> str:
    return data.capitalize()


@register.filter
def whitespace_char(data: str, char: str) -> str:
    return data.replace(char, ' ')


@register.filter
def split(data: str, split_at: str) -> list:
    return data.split(split_at)


@register.filter
def find(data: str, search: str) -> bool:
    if not isinstance(data, str):
        data = str(data)

    return data.find(search) != -1


# see: https://github.com/grafana/django-saml2-auth/blob/main/django_saml2_auth/errors.py
SAML_ERROR_CODES = {
    1100: 'EMPTY_FUNCTION_PATH',
    1101: 'PATH_ERROR',
    1102: 'IMPORT_ERROR',
    1103: 'GENERAL_EXCEPTION',
    1104: 'CREATE_USER_ERROR',
    1105: 'GROUP_JOIN_ERROR',
    1106: 'NO_REVERSE_MATCH',
    1107: 'ERROR_CREATING_SAML_CONFIG_OR_CLIENT',
    1108: 'NO_SAML_RESPONSE_FROM_CLIENT',
    1109: 'NO_SAML_RESPONSE_FROM_IDP',
    1110: 'NO_NAME_ID_IN_SAML_RESPONSE',
    1111: 'NO_ISSUER_IN_SAML_RESPONSE',
    1112: 'NO_USER_IDENTITY_IN_SAML_RESPONSE',
    1113: 'NO_TOKEN_SPECIFIED',
    1114: 'NO_USERNAME_OR_EMAIL_SPECIFIED',
    1115: 'SHOULD_NOT_CREATE_USER',
    1116: 'INACTIVE_USER',
    1117: 'NO_METADATA_URL_OR_FILE',
    1118: 'NO_SAML_CLIENT',
    1119: 'NO_JWT_ALGORITHM',
    1120: 'INVALID_METADATA_URL',
    1121: 'NO_METADATA_URL_ASSOCIATED',
    1122: 'INVALID_REQUEST_METHOD',
    1123: 'CANNOT_DECODE_JWT_TOKEN',
    1124: 'USER_MISMATCH',
    1125: 'NO_JWT_SECRET',
    1126: 'NO_JWT_PRIVATE_KEY',
    1127: 'NO_JWT_PUBLIC_KEY',
    1128: 'INVALID_JWT_ALGORITHM',
    1129: 'NO_USER_ID',
    1130: 'INVALID_TOKEN',
    1131: 'INVALID_NEXT_URL',
}


@register.filter
def saml_error_by_code(error_code: int) -> str:
    if error_code not in SAML_ERROR_CODES:
        return ''

    return f" ({SAML_ERROR_CODES[error_code]})"



from django.shortcuts import render
from django.shortcuts import HttpResponse
from django.urls import path
from django.contrib.auth.decorators import login_required

from aw.utils.http import ui_endpoint_wrapper, ui_endpoint_wrapper_kwargs
from aw.model.job import JobExecution, JobExecutionResultHost
from aw.api_endpoints.job_util import get_viewable_jobs
from aw.utils.util import get_next_cron_execution_str
from aw.views.forms.job import job_edit, job_clone, job_credentials_edit, job_repository_static_edit, \
    job_repository_git_edit

LIMIT_JOB_RESULTS = 10
LIMIT_JOB_LOG_RESULTS = 50


@login_required
@ui_endpoint_wrapper
def manage(request) -> HttpResponse:
    jobs_viewable = get_viewable_jobs(request.user)
    executions = {}
    next_executions = {}
    execution_results_hosts = {}

    for job in jobs_viewable:
        executions[job.id] = JobExecution.objects.filter(job=job).order_by('-updated')[:LIMIT_JOB_RESULTS]

        cron = '-'
        try:
            if job.schedule is not None:
                cron = get_next_cron_execution_str(job.schedule)

        except ValueError:
            pass

        next_executions[job.id] = cron

        for execution in executions[job.id]:
            if execution.result is not None:
                execution_results_hosts[execution.id] = JobExecutionResultHost.objects.filter(
                    result=execution.result,
                ).order_by('-created')

    return render(
        request, status=200, template_name='jobs/manage.html',
        context={
            'jobs': jobs_viewable, 'executions': executions, 'next_executions': next_executions,
            'show_update_time': True, 'execution_results_hosts': execution_results_hosts,
        }
    )


@login_required
@ui_endpoint_wrapper_kwargs
def job_logs(request) -> HttpResponse:
    return render(request, status=200, template_name='jobs/logs.html', context={'show_update_time': True})


@login_required
@ui_endpoint_wrapper_kwargs
def job_credentials(request) -> HttpResponse:
    return render(request, status=200, template_name='jobs/credentials.html', context={'show_update_time': True})


@login_required
@ui_endpoint_wrapper_kwargs
def job_repository(request) -> HttpResponse:
    return render(request, status=200, template_name='jobs/repository.html', context={'show_update_time': True})


urlpatterns_jobs = [
    path('ui/jobs/credentials/<int:credentials_id>', job_credentials_edit),
    path('ui/jobs/credentials', job_credentials),
    path('ui/jobs/log', job_logs),
    path('ui/jobs/manage/job/clone/<int:job_id>', job_clone),
    path('ui/jobs/manage/job/<int:job_id>', job_edit),
    path('ui/jobs/manage/job', job_edit),
    path('ui/jobs/repository/static/<int:repo_id>', job_repository_static_edit),
    path('ui/jobs/repository/git/<int:repo_id>', job_repository_git_edit),
    path('ui/jobs/repository', job_repository),
    path('ui/jobs/manage', manage),
]



from django.contrib import admin
from rest_framework_api_key.admin import APIKey

from aw.model.api import AwAPIKey
from aw.model.job import Job, JobExecution, JobExecutionResult, JobError, JobExecutionResultHost
from aw.model.permission import JobPermission, JobPermissionMemberUser, JobPermissionMemberGroup, \
    JobPermissionMapping
from aw.model.job_credential import JobGlobalCredentials, JobUserCredentials


admin.site.unregister(APIKey)

admin.site.register(Job)
admin.site.register(JobExecution)
admin.site.register(JobPermission)
admin.site.register(JobPermissionMemberUser)
admin.site.register(JobPermissionMemberGroup)
admin.site.register(JobPermissionMapping)
admin.site.register(JobExecutionResult)
admin.site.register(JobExecutionResultHost)
admin.site.register(JobError)
admin.site.register(JobGlobalCredentials)
admin.site.register(JobUserCredentials)
admin.site.register(AwAPIKey)



from functools import cache

from sys import version_info
from pathlib import Path
from datetime import datetime
from collections import OrderedDict
from importlib import metadata

from aw.config.main import VERSION
from aw.utils.subps import process_cache
from aw.utils.util import datetime_from_db_str
from aw.config.hardcoded import LOG_TIME_FORMAT
from aw.utils.deployment import deployment_docker
from aw.model.system import get_schema_metadata


@cache
def get_version() -> str:
    if VERSION == 'latest' or VERSION.find('dev') != -1 or VERSION.find('staging') != -1:
        this_file = Path(__file__)
        repo_base = this_file.resolve().parent.parent.parent.parent.parent
        if (repo_base / '.git').is_dir():
            commit = process_cache(cmd='git rev-parse --short HEAD', cwd=repo_base)['stdout'].strip()
            if commit != '':
                return f'{VERSION} ({commit})'

        else:
            mod_time = this_file.stat().st_mtime
            mod_time = datetime.fromtimestamp(mod_time).strftime(LOG_TIME_FORMAT)
            return f'{VERSION} ({mod_time})'

    return VERSION


def parsed_ansible_version(python_modules) -> dict:
    versions = {'ansible': None, 'jinja': None, 'libyaml': None, 'ansible_runner': None}
    try:
        ansible_version = process_cache('ansible --version')['stdout'].split('\n')
        versions['ansible_core'] = ansible_version[0].strip()
        versions['jinja'] = ansible_version[-2].split('=')[1].strip()
        versions['libyaml'] = ansible_version[-1].split('=')[1].strip()

        if 'ansible-runner' in python_modules:
            versions['ansible_runner'] = python_modules['ansible-runner']['version']

        if 'ansible' in python_modules:
            versions['ansible'] = python_modules['ansible']['version']

    except IndexError:
        pass

    return versions


def parsed_python_modules() -> dict:
    modules = OrderedDict()
    try:
        module_list = [m[0] for m in metadata.packages_distributions().values()]

        for module in sorted(module_list):
            modules[module.lower()] = {'name': module, 'version': metadata.distribution(module).version}

    except (ImportError, AttributeError):
        result = process_cache('pip list')
        if result['rc'] != 0:
            return {}

        for line in result['stdout'].split('\n'):
            if line.find('.') == -1:
                continue

            name, version = line.split(' ', 1)
            name = name.strip()
            modules[name.lower()] = {'name': name, 'version': version.strip()}

    return modules


def get_system_versions(python_modules: dict = None, ansible_version: dict = None) -> dict:
    if python_modules is None:
        python_modules = parsed_python_modules()

    if ansible_version is None:
        ansible_version = parsed_ansible_version(python_modules)

    linux_versions = process_cache('uname -a')['stdout']
    if deployment_docker():
        linux_versions += ' (dockerized)'

    db_schema = get_schema_metadata()

    return {
        'env_linux': linux_versions,
        'env_db_schema': f'{db_schema.schema_version} (updated: {datetime_from_db_str(db_schema.updated)})',
        'env_git': process_cache('git --version')['stdout'],
        'env_ansible_core': ansible_version['ansible_core'],
        'env_ansible_runner': ansible_version['ansible_runner'],
        'env_django': python_modules['django']['version'],
        'env_django_api': python_modules['djangorestframework']['version'],
        'env_gunicorn': python_modules['gunicorn']['version'],
        'env_jinja': ansible_version['jinja'],
        'env_libyaml': ansible_version['libyaml'],
        'env_python': f"{version_info.major}.{version_info.minor}.{version_info.micro}",
    }



from urllib.parse import urljoin

from django.shortcuts import redirect, render, HttpResponse
from django_saml2_auth.user import create_jwt_token

from aw.utils.http import ui_endpoint_wrapper_auth
from aw.settings import SAML2_AUTH, LOGIN_PATH, LOGIN_REDIRECT_URL


# SP-initiated SAML SSO; see: https://github.com/grafana/django-saml2-auth/issues/105
@ui_endpoint_wrapper_auth
def saml_sp_initiated_login(request) -> HttpResponse:
    if request.user.is_authenticated:
        return redirect(LOGIN_REDIRECT_URL)

    return render(request, status=200, template_name='registration/saml.html')


@ui_endpoint_wrapper_auth
def saml_sp_initiated_login_init(request) -> HttpResponse:
    if request.user.is_authenticated:
        return redirect(LOGIN_REDIRECT_URL)

    if request.method != 'POST' or 'username' not in request.POST:
        return redirect(f"{LOGIN_PATH}?error=Required 'username' was not provided!")

    token = create_jwt_token(request.POST['username'])
    assertion_url = SAML2_AUTH['ASSERTION_URL']
    sso_init_url = urljoin(assertion_url, f'a/saml/sp/?token={token}')
    return redirect(sso_init_url)



from django.db import models
from rest_framework_api_key.models import AbstractAPIKey

from aw.base import USERS


class AwAPIKey(AbstractAPIKey):
    user = models.ForeignKey(USERS, on_delete=models.CASCADE, editable=False)



from django.urls import path
from django.shortcuts import HttpResponse
from django.shortcuts import render
from django.contrib.auth.decorators import login_required

from aw.utils.http import ui_endpoint_wrapper
from aw.views.forms.settings import setting_permission_edit


@login_required
@ui_endpoint_wrapper
def setting_api_key(request) -> HttpResponse:
    return render(request, status=200, template_name='settings/api_key.html', context={'show_update_time': True})


@login_required
@ui_endpoint_wrapper
def setting_permission(request) -> HttpResponse:
    return render(request, status=200, template_name='settings/permission.html', context={'show_update_time': True})


urlpatterns_settings = [
    path('ui/settings/api_keys', setting_api_key),
    path('ui/settings/permissions/<int:perm_id>', setting_permission_edit),
    path('ui/settings/permissions', setting_permission),
]



from multiprocessing import cpu_count
from string import ascii_letters
from random import choice as random_choice
from pathlib import Path
from ssl import PROTOCOL_TLSv1_2

import gunicorn
from gunicorn.app.wsgiapp import WSGIApplication

from aw.utils.deployment import deployment_dev, deployment_docker
from aw.utils.debug import log, warn_if_development
from aw.config.environment import get_aw_env_var_or_default

PORT_WEB = get_aw_env_var_or_default('port')
LISTEN_ADDRESS = get_aw_env_var_or_default('address')

# https://docs.gunicorn.org/en/stable/settings.html
OPTIONS_DEV = {
    'reload': True,
    'loglevel': 'info',
    'workers': 2,
}
OPTIONS_PROD = {
    'bind': f'{LISTEN_ADDRESS}:{PORT_WEB}',
    'reload': False,
    'loglevel': 'warning',
}

if deployment_docker():
    OPTIONS_PROD['bind'] = f'0.0.0.0:{PORT_WEB}'


class StandaloneApplication(WSGIApplication):
    def __init__(self, app_uri, options=None):
        self.options = options or {}
        self.app_uri = app_uri
        super().__init__()

    def load_config(self):
        config = {
            key: value
            for key, value in self.options.items()
            if key in self.cfg.settings and value is not None
        }
        for key, value in config.items():
            self.cfg.set(key.lower(), value)


def init_webserver():
    gunicorn.SERVER = ''.join(random_choice(ascii_letters) for _ in range(10))
    opts = {
        'workers': (cpu_count() * 2) + 1,
        **OPTIONS_PROD
    }
    if deployment_dev():
        warn_if_development()
        opts = {**opts, **OPTIONS_DEV}

    scheme = 'http'
    ssl_cert = get_aw_env_var_or_default('ssl_file_crt')
    ssl_key = get_aw_env_var_or_default('ssl_file_key')
    if ssl_cert is not None and ssl_key is not None:
        if not Path(ssl_cert).is_file() or not Path(ssl_key).is_file():
            log(
                msg=f"Either SSL certificate or SSL key is not readable: {ssl_cert}, {ssl_key}",
                level=1,
            )

        else:
            opts = {
                **opts,
                'keyfile': ssl_key,
                'certfile': ssl_cert,
                'ssl_version': PROTOCOL_TLSv1_2,
                'do_handshake_on_connect': True,
            }
            scheme = 'https'

    log(msg=f"Listening on {scheme}://{opts['bind']}", level=5)

    StandaloneApplication(
        app_uri="aw.main:app",
        options=opts
    ).run()



from os import listdir
from pathlib import Path
from functools import cache

from django.core.exceptions import ObjectDoesNotExist
from rest_framework.views import APIView
from rest_framework import serializers
from rest_framework.response import Response
from drf_spectacular.utils import extend_schema, OpenApiResponse, OpenApiParameter

from aw.config.main import config
from aw.api_endpoints.base import API_PERMISSION, BaseResponse, GenericResponse
from aw.model.repository import Repository
from aw.execute.repository import get_path_repo_wo_isolate
from aw.utils.util import is_set


class FileSystemReadResponse(BaseResponse):
    files = serializers.ListSerializer(child=serializers.CharField())
    directories = serializers.ListSerializer(child=serializers.CharField())


class FileSystemExistsResponse(BaseResponse):
    exists = serializers.BooleanField()
    fstype = serializers.CharField()


class APIFsBrowse(APIView):
    http_method_names = ['get']
    serializer_class = FileSystemReadResponse
    permission_classes = API_PERMISSION

    @staticmethod
    @cache
    def _listdir(path: str) -> list[str]:
        return listdir(path)

    @classmethod
    @extend_schema(
        request=None,
        responses={
            200: FileSystemReadResponse,
            400: OpenApiResponse(GenericResponse, description='Invalid browse-selector provided'),
            403: OpenApiResponse(GenericResponse, description='Traversal not allowed'),
            404: OpenApiResponse(GenericResponse, description='Base directory does not exist'),
        },
        summary='Return list of existing files and directories.',
        description="This endpoint is mainly used for form auto-completion when selecting job-files",
        parameters=[
            OpenApiParameter(
                name='base', type=str, default='/', description='Relative directory to query',
                required=False,
            ),
        ],
    )
    def get(cls, request, repository: int = None):
        # pylint: disable=R0912
        browse_root = Path(config['path_play'])
        items = {'files': [], 'directories': []}

        if repository not in [None, 0, '0']:
            try:
                repository = Repository.objects.get(id=repository)
                if repository is None:
                    raise ObjectDoesNotExist

                if repository.rtype_name == 'Static':
                    browse_root = repository.static_path

                else:
                    if repository.git_isolate:
                        # do not validate as the repo does not exist..
                        all_valid = ['.*']
                        items['files'] = all_valid
                        items['directories'] = all_valid
                        return Response(items)

                    browse_root = get_path_repo_wo_isolate(repository)
                    if is_set(repository.git_playbook_base):
                        browse_root = browse_root / repository.git_playbook_base

            except ObjectDoesNotExist:
                return Response(data={'msg': 'Provided repository does not exist'}, status=404)

        if not browse_root.is_dir():
            return Response(data={'msg': f"Base directory '{browse_root}' does not exist"}, status=404)

        if 'base' not in request.GET:
            base = '/'
        else:
            base = str(request.GET['base'])

        if base.find('..') != -1 or base.startswith('/'):
            return Response(data={'msg': 'Traversal not allowed'}, status=403)

        browse_root = browse_root / base

        if not browse_root.is_dir():
            return Response(data={'msg': f"Base directory '{browse_root}' does not exist"}, status=404)

        raw_items = cls._listdir(browse_root)

        for item in raw_items:
            try:
                item_path = browse_root / item
                if item_path.is_file():
                    items['files'].append(item)
                elif item_path.is_dir():
                    items['directories'].append(item)

            except OSError:
                continue

        return Response(items)


class APIFsExists(APIView):
    http_method_names = ['get']
    serializer_class = FileSystemExistsResponse
    permission_classes = API_PERMISSION

    @classmethod
    @extend_schema(
        request=None,
        responses={
            200: FileSystemExistsResponse,
            400: OpenApiResponse(GenericResponse, description='No file or directory not provided'),
            403: OpenApiResponse(GenericResponse, description='Access to file or directory is forbidden'),
        },
        summary='Return if the provided file or directory exists.',
        description="This endpoint is mainly used for form validation when configuring path and files",
        parameters=[
            OpenApiParameter(
                name='item', type=str, default=None, required=True,
                description='File or directory that should be checked',
            ),
        ],
    )
    def get(cls, request):
        if 'item' not in request.GET:
            return Response(data={'msg': "Required parameter 'item' was not provided"}, status=400)

        try:
            fs_item = Path(request.GET['item'])
            try:
                return Response({
                    'exists': fs_item.exists(),
                    'fstype': 'file' if fs_item.is_file() else 'directory',
                })

            except OSError:
                return Response({'exists': False, 'fstype': 'unknown'})

        except PermissionError:
            return Response(data={'msg': 'Access to file or directory is forbidden'}, status=403)



from pathlib import Path

from django.shortcuts import HttpResponse
from rest_framework import serializers

from aw.config.main import config
from aw.config.hardcoded import JOB_EXECUTION_LIMIT
from aw.model.job import Job, JobExecution
from aw.utils.permission import get_viewable_jobs
from aw.utils.util import get_next_cron_execution_str
from aw.base import USERS


class JobReadResponse(serializers.ModelSerializer):
    class Meta:
        model = Job
        fields = Job.api_fields_read

    next_run = serializers.CharField(required=False)


class JobExecutionReadResponse(serializers.ModelSerializer):
    class Meta:
        model = JobExecution
        fields = JobExecution.api_fields_read

    job_name = serializers.CharField(required=False)
    job_comment = serializers.CharField(required=False)
    user_name = serializers.CharField(required=False)
    status_name = serializers.CharField(required=False)
    failed = serializers.BooleanField(required=False)
    error_s = serializers.CharField(required=False)
    error_m = serializers.CharField(required=False)
    time_start = serializers.CharField(required=False)
    time_fin = serializers.CharField(required=False)
    log_stdout = serializers.CharField(required=False)
    log_stdout_url = serializers.CharField(required=False)
    log_stderr = serializers.CharField(required=False)
    log_stderr_url = serializers.CharField(required=False)
    log_stderr_repo_url = serializers.CharField(required=False)
    log_stdout_repo = serializers.CharField(required=False)
    log_stdout_repo_url = serializers.CharField(required=False)
    log_stderr_repo = serializers.CharField(required=False)


def get_job_execution_serialized(execution: JobExecution) -> dict:
    serialized = JobExecutionReadResponse(instance=execution).data
    serialized['job'] = execution.job.id
    serialized['job_name'] = execution.job.name
    serialized['job_comment'] = execution.job.comment
    serialized['user'] = execution.user.id if execution.user is not None else None
    serialized['user_name'] = execution.user.username if execution.user is not None else 'Scheduled'
    serialized['time_start'] = execution.time_created_str
    serialized['time_fin'] = None
    serialized['failed'] = None
    serialized['error_s'] = None
    serialized['error_m'] = None

    for logfile in JobExecution.log_file_fields:
        if serialized[logfile] is None or not Path(serialized[logfile]).is_file():
            serialized[logfile] = None
            serialized[logfile + '_url'] = None

    if execution.result is not None:
        serialized['time_fin'] = execution.result.time_fin_str
        serialized['failed'] = execution.result.failed
        if execution.result.error is not None:
            serialized['error_s'] = execution.result.error.short
            serialized['error_m'] = execution.result.error.med

    return serialized


def get_job_executions_serialized(job: Job, execution_count: int = JOB_EXECUTION_LIMIT) -> list[dict]:
    serialized = []
    for execution in JobExecution.objects.filter(job=job).order_by('-updated')[:execution_count]:
        serialized.append(get_job_execution_serialized(execution))

    return serialized


def get_viewable_jobs_serialized(
        user: USERS, executions: bool = False,
        execution_count: int = None
) -> list[dict]:
    serialized = []

    for job in get_viewable_jobs(user):
        job_serialized = JobReadResponse(instance=job).data
        job_serialized['next_run'] = None

        try:
            if job.schedule is not None and job.enabled:
                job_serialized['next_run'] = get_next_cron_execution_str(job.schedule) + f" {config['timezone']}"

        except ValueError:
            pass

        if executions:
            job_serialized['executions'] = get_job_executions_serialized(job=job, execution_count=execution_count)

        serialized.append(job_serialized)

    return serialized


def get_log_file_content(logfile: (str, Path)) -> HttpResponse:
    with open(logfile, 'rb') as _logfile:
        content_b = _logfile.read()
        if content_b == b'':
            return HttpResponse(content_b, content_type='text/plain', status=404)

        response = HttpResponse(content_b, content_type='text/plain', status=200)

    response['Content-Disposition'] = f"inline; filename={logfile.rsplit('/', 1)[1]}"
    return response



from threading import Thread
from pathlib import Path

from django.core.exceptions import ObjectDoesNotExist
from django.db.utils import IntegrityError
from rest_framework.generics import GenericAPIView
from rest_framework import serializers
from rest_framework.response import Response
from drf_spectacular.utils import extend_schema, OpenApiResponse, OpenApiParameter

from aw.model.repository import Repository
from aw.api_endpoints.base import API_PERMISSION, GenericResponse, get_api_user, LogDownloadResponse
from aw.utils.permission import has_manager_privileges, has_repository_permission, get_viewable_repositories
from aw.model.job import Job
from aw.utils.util import unset_or_null, is_set
from aw.model.permission import CHOICE_PERMISSION_READ, CHOICE_PERMISSION_WRITE, CHOICE_PERMISSION_DELETE, \
    CHOICE_PERMISSION_EXECUTE
from aw.execute.repository import api_update_repository
from aw.api_endpoints.job_util import get_log_file_content


class RepositoryWriteRequest(serializers.ModelSerializer):
    class Meta:
        model = Repository
        fields = Repository.api_fields_write


class RepositoryReadResponse(RepositoryWriteRequest):
    class Meta:
        model = Repository
        fields = Repository.api_fields_read

    rtype_name = serializers.CharField()
    status_name = serializers.CharField()
    log_stdout_url = serializers.CharField()
    log_stderr_url = serializers.CharField()


def repository_in_use(repository: Repository) -> bool:
    return Job.objects.filter(repository=repository).exists()


def validate_repository_types(repository: dict) -> (bool, str):
    rtype_name = Repository.rtype_name_from_id(repository['rtype'])
    if rtype_name == 'Git':
        try:
            if is_set(repository['git_override_initialize']) and is_set(repository['git_override_update']):
                return True, ''

        except KeyError:
            pass

        if unset_or_null(repository, 'git_origin'):
            return False, 'Git Origin is required'

        if unset_or_null(repository, 'git_branch'):
            return False, 'Git Branch is required'

    else:
        if unset_or_null(repository, 'static_path'):
            return False, 'Static Path is required'

    return True, ''


def build_repository(repository: Repository) -> dict:
    data = RepositoryReadResponse(instance=repository).data
    data['time_update'] = repository.time_update_str
    if data['log_stderr'] is None or not Path(data['log_stderr']).is_file():
        data['log_stderr'] = None
        data['log_stderr_url'] = None

    return data


class APIRepository(GenericAPIView):
    http_method_names = ['get', 'post']
    serializer_class = RepositoryReadResponse
    permission_classes = API_PERMISSION

    @staticmethod
    @extend_schema(
        request=None,
        responses={200: RepositoryReadResponse},
        summary='Return list of repositories',
        operation_id='repository_list',
    )
    def get(request):
        user = get_api_user(request)
        repositories = []

        for repository in get_viewable_repositories(user=user):
            repositories.append(build_repository(repository))

        return Response(data=repositories, status=200)

    @extend_schema(
        request=RepositoryWriteRequest,
        responses={
            200: OpenApiResponse(response=GenericResponse, description='Repository created'),
            400: OpenApiResponse(response=GenericResponse, description='Invalid repository data provided'),
            403: OpenApiResponse(response=GenericResponse, description='Not privileged to create a repository'),
        },
        summary='Create a new Repository.',
        operation_id='repository_create',
    )
    def post(self, request):
        privileged = has_manager_privileges(user=get_api_user(request), kind='repository')
        if not privileged:
            return Response(data={'msg': 'Not privileged to manage repositories'}, status=403)

        serializer = RepositoryWriteRequest(data=request.data)

        if not serializer.is_valid():
            return Response(data={'msg': f"Provided repository data is not valid: '{serializer.errors}'"}, status=400)

        rtype_valid, rtype_error = validate_repository_types(serializer.validated_data)
        if not rtype_valid:
            return Response(data={'msg': f"Provided repository data is not valid: '{rtype_error}'"}, status=400)

        try:
            serializer.save()

        except IntegrityError as err:
            return Response(data={'msg': f"Provided repository data is not valid: '{err}'"}, status=400)

        return Response({'msg': f"Repository '{serializer.data['name']}' created successfully"})


class APIRepositoryItem(GenericAPIView):
    http_method_names = ['get', 'put', 'post', 'delete']
    serializer_class = GenericResponse
    permission_classes = API_PERMISSION

    @staticmethod
    @extend_schema(
        request=None,
        responses={
            200: RepositoryReadResponse,
            404: OpenApiResponse(response=GenericResponse, description='Repository does not exist'),
        },
        summary='Return information of a repository.',
        operation_id='repository_get'
    )
    def get(request, repo_id: int):
        user = get_api_user(request)

        try:
            repository = Repository.objects.get(id=repo_id)
            if not has_repository_permission(
                    user=user,
                    repository=repository,
                    permission_needed=CHOICE_PERMISSION_READ
            ):
                return Response(
                    data={'msg': f"Not privileged to view the repository '{repository.name}'"},
                    status=403,
                )

            return Response(data=build_repository(repository), status=200)

        except ObjectDoesNotExist:
            return Response(data={'msg': f"Repository with ID {repo_id} does not exist"}, status=404)

    @extend_schema(
        request=RepositoryWriteRequest,
        responses={
            200: OpenApiResponse(response=GenericResponse, description='Repository updated'),
            400: OpenApiResponse(response=GenericResponse, description='Invalid repository data provided'),
            403: OpenApiResponse(response=GenericResponse, description='Not privileged to edit the repository'),
            404: OpenApiResponse(response=GenericResponse, description='Repository does not exist'),
        },
        summary='Modify a repository.',
        operation_id='repository_edit',
    )
    def put(self, request, repo_id: int):
        user = get_api_user(request)

        serializer = RepositoryWriteRequest(data=request.data)
        if not serializer.is_valid():
            return Response(data={'msg': f"Provided repository data is not valid: '{serializer.errors}'"}, status=400)

        rtype_valid, rtype_error = validate_repository_types(serializer.validated_data)
        if not rtype_valid:
            return Response(data={'msg': f"Provided repository data is not valid: '{rtype_error}'"}, status=400)

        try:
            repository = Repository.objects.get(id=repo_id)

        except ObjectDoesNotExist:
            repository = None

        if repository is None:
            return Response(data={'msg': f"Repository with ID {repo_id} does not exist"}, status=404)

        if not has_repository_permission(
                user=user,
                repository=repository,
                permission_needed=CHOICE_PERMISSION_WRITE
        ):
            return Response(
                data={'msg': f"Not privileged to modify the repository '{repository.name}'"},
                status=403,
            )

        try:
            Repository.objects.filter(id=repo_id).update(**serializer.data)
            return Response(data={'msg': f"Repository '{repository.name}' updated"}, status=200)

        except IntegrityError as err:
            return Response(data={'msg': f"Provided repository data is not valid: '{err}'"}, status=400)

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(response=GenericResponse, description='Repository deleted'),
            400: OpenApiResponse(response=GenericResponse, description='Invalid repository data provided'),
            403: OpenApiResponse(response=GenericResponse, description='Not privileged to delete the repository'),
            404: OpenApiResponse(response=GenericResponse, description='Repository does not exist'),
        },
        summary='Delete a repository.',
        operation_id='repository_delete'
    )
    def delete(self, request, repo_id: int):
        user = get_api_user(request)

        try:
            repository = Repository.objects.get(id=repo_id)
            if repository is not None:
                if not has_repository_permission(
                        user=user,
                        repository=repository,
                        permission_needed=CHOICE_PERMISSION_DELETE
                ):
                    return Response(
                        data={'msg': f"Not privileged to delete the repository '{repository.name}'"},
                        status=403,
                    )

                if repository_in_use(repository):
                    return Response(
                        data={'msg': f"Repository '{repository.name}' cannot be deleted as it is still in use"},
                        status=400,
                    )

                repository.delete()
                return Response(data={'msg': f"Repository '{repository.name}' deleted"}, status=200)

        except ObjectDoesNotExist:
            pass

        return Response(data={'msg': f"Repository with ID {repo_id} does not exist"}, status=404)

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(response=GenericResponse, description='Repository update initiated'),
            403: OpenApiResponse(response=GenericResponse, description='Not privileged to update the repository'),
            404: OpenApiResponse(response=GenericResponse, description='Repository does not exist'),
        },
        summary='Download/Update a repository.',
        operation_id='repository_update'
    )
    def post(self, request, repo_id: int):
        user = get_api_user(request)

        try:
            repository = Repository.objects.get(id=repo_id)
            if repository is not None:
                if not has_repository_permission(
                        user=user,
                        repository=repository,
                        permission_needed=CHOICE_PERMISSION_EXECUTE,
                ):
                    return Response(
                        data={'msg': f"Not privileged to update the repository '{repository.name}'"},
                        status=403,
                    )

                repository_update_thread = Thread(
                    target=api_update_repository,
                    kwargs={'repository': repository, 'user': user}
                )
                repository_update_thread.start()

                return Response(data={'msg': f"Repository '{repository.name}' update initiated"}, status=200)

        except ObjectDoesNotExist:
            pass

        return Response(data={'msg': f"Repository with ID {repo_id} does not exist"}, status=404)


class APIRepositoryLogFile(GenericAPIView):
    http_method_names = ['get']
    serializer_class = LogDownloadResponse
    permission_classes = API_PERMISSION
    valid_logfile_type = ['stdout', 'stderr']

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(GenericResponse, description='Download repository log-file'),
            403: OpenApiResponse(GenericResponse, description='Not privileged to view the repository logs'),
            404: OpenApiResponse(GenericResponse, description='Repository or log-file dos not exist'),
        },
        summary='Download log-file of the last repository update.',
        operation_id='repository_logfile',
        parameters=[
            OpenApiParameter(
                name='type', type=str, default='stdout',
                description=f"Type of log-file to download. One of {valid_logfile_type}",
                required=False,
            ),
        ],
    )
    def get(self, request, repo_id: int):
        user = get_api_user(request)
        try:
            repository = Repository.objects.get(id=repo_id)
            if repository is not None:
                if not has_repository_permission(
                        user=user,
                        repository=repository,
                        permission_needed=CHOICE_PERMISSION_READ,
                ):
                    return Response(
                        data={'msg': f"Not privileged to view logs of the repository '{repository.name}'"},
                        status=403,
                    )

                logfile = repository.log_stdout
                if 'type' in request.GET:
                    logfile_type = request.GET['type'] if request.GET['type'] in self.valid_logfile_type else 'stdout'
                    logfile = getattr(repository, f'log_{logfile_type}')

                if logfile is None:
                    return Response(data={'msg': f"No logs found for repository '{repository.name}'"}, status=404)

                return get_log_file_content(logfile)

        except (ObjectDoesNotExist, FileNotFoundError):
            pass

        return Response(
            data={'msg': f"Repository with ID '{repo_id}' or log-file does not exist"},
            status=404,
        )



#!/usr/bin/env python3

from sys import argv as sys_argv
from sys import path as sys_path
from os import path as os_path

# pylint: disable=C0415


def main():
    # pylint: disable=E0401,C0415
    try:
        from cli_init import init_cli

    except ModuleNotFoundError:
        sys_path.append(os_path.dirname(os_path.abspath(__file__)))
        from cli_init import init_cli

    init_cli()
    from django.core.management import execute_from_command_line
    execute_from_command_line(sys_argv)


if __name__ == '__main__':
    main()



import traceback

from ansible_runner import RunnerConfig, Runner

from aw.config.main import config
from aw.model.job import Job, JobExecution, JobExecutionResult
from aw.execute.play_util import runner_cleanup, runner_prep, parse_run_result, failure, runner_logs
from aw.execute.util import get_path_run, is_execution_status, job_logs
from aw.execute.repository import ExecuteRepository
from aw.utils.util import datetime_w_tz, is_null, timed_lru_cache  # get_ansible_versions
from aw.utils.handlers import AnsibleConfigError
from aw.utils.debug import log


class AwRunnerConfig(RunnerConfig):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.quiet = True
        self.runner_mode = 'subprocess'
        self.timeout = config['run_timeout']


def ansible_playbook(job: Job, execution: (JobExecution, None)):
    time_start = datetime_w_tz()
    path_run = get_path_run()
    if is_null(execution):
        execution = JobExecution(user=None, job=job, comment='Scheduled')

    result = JobExecutionResult(time_start=time_start)
    result.save()
    execution.result = result
    execution.save()

    log_files = job_logs(job=job, execution=execution)

    @timed_lru_cache(seconds=1)  # check actual status every N seconds; lower DB queries
    def _cancel_job() -> bool:
        return is_execution_status(execution, 'Stopping')

    exec_repo = ExecuteRepository(repository=job.repository, execution=execution, path_run=path_run)
    try:
        exec_repo.create_or_update_repository()
        project_dir = exec_repo.get_project_dir()
        opts = runner_prep(job=job, execution=execution, path_run=path_run, project_dir=project_dir)
        execution.save()

        runner_cfg = AwRunnerConfig(**opts)
        runner_logs(cfg=runner_cfg, log_files=log_files)
        runner_cfg.prepare()
        command = ' '.join(runner_cfg.command)
        log(msg=f"Running job '{job.name}': '{command}'", level=5)
        execution.command = command
        execution.save()

        runner = Runner(config=runner_cfg, cancel_callback=_cancel_job)
        runner.run()

        parse_run_result(
            result=result,
            execution=execution,
            runner=runner,
        )
        del runner

        runner_cleanup(execution=execution, path_run=path_run, exec_repo=exec_repo)

    except (OSError, AnsibleConfigError) as err:
        tb = traceback.format_exc(limit=1024)
        failure(
            execution=execution, exec_repo=exec_repo, path_run=path_run, result=result,
            error_s=str(err), error_m=tb,
        )
        raise



from os import getpid
from sys import stderr, stdout
from inspect import stack as inspect_stack
from inspect import getfile as inspect_getfile

from aw.config.main import config
from aw.utils.util import datetime_w_tz
from aw.utils.deployment import deployment_dev, deployment_staging
from aw.config.hardcoded import LOG_TIME_FORMAT

PID = getpid()

LEVEL_NAME_MAPPING = {
    1: 'FATAL',
    2: 'ERROR',
    3: 'WARN',
    4: 'INFO',
    5: 'INFO',
    6: 'DEBUG',
    7: 'DEBUG',
}


def _log_prefix() -> str:
    # time format adapted to the one used by gunicorn
    # todo: update gunicorn log format (gunicorn.glogging.CONFIG_DEFAULTS)
    return f'[{datetime_w_tz().strftime(LOG_TIME_FORMAT)}] [{PID}]'


def log(msg: str, level: int = 3):
    debug = deployment_dev() or config['debug']
    prefix_caller = ''

    if level > 5 and not debug:
        return

    if debug:
        caller = inspect_getfile(inspect_stack()[1][0]).rsplit('/', 1)[1].rsplit('.', 1)[0]
        prefix_caller = f'[{caller}] '

    print(f"{_log_prefix()} [{LEVEL_NAME_MAPPING[level]}] {prefix_caller}{msg}")


def log_warn(msg: str, _stderr: bool = False):
    if _stderr:
        stderr.write(f'\x1b[1;33m{_log_prefix()} [{LEVEL_NAME_MAPPING[3]}] {msg}\x1b[0m\n')

    else:
        stdout.write(f'\x1b[1;33m{_log_prefix()} [{LEVEL_NAME_MAPPING[3]}] {msg}\x1b[0m\n')


def log_error(msg: str):
    stderr.write(f'\033[01;{_log_prefix()} [{LEVEL_NAME_MAPPING[2]}] {msg}\x1b[0m\n')


def warn_if_development():
    if deployment_dev():
        log_warn('Development mode!')

    elif deployment_staging():
        log_warn('Staging mode!')



from django.core.exceptions import ObjectDoesNotExist
from django.db.utils import IntegrityError
from rest_framework.views import APIView
from rest_framework import serializers
from rest_framework.response import Response
from drf_spectacular.utils import extend_schema, OpenApiResponse, OpenApiParameter

from aw.model.job import Job, JobExecution
from aw.model.job_credential import BaseJobCredentials, JobUserCredentials, JobGlobalCredentials
from aw.model.permission import CHOICE_PERMISSION_READ, CHOICE_PERMISSION_WRITE, CHOICE_PERMISSION_DELETE
from aw.api_endpoints.base import API_PERMISSION, get_api_user, GenericResponse, BaseResponse
from aw.utils.permission import has_credentials_permission, has_manager_privileges
from aw.utils.util import is_null
from aw.base import USERS


class JobGlobalCredentialsReadResponse(serializers.ModelSerializer):
    class Meta:
        model = JobGlobalCredentials
        fields = JobGlobalCredentials.api_fields_read

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        for secret_attr in BaseJobCredentials.SECRET_ATTRS:
            setattr(self, f'{secret_attr}_is_set', serializers.BooleanField(required=False))


class JobUserCredentialsReadResponse(JobGlobalCredentialsReadResponse):
    class Meta:
        model = JobUserCredentials
        fields = JobUserCredentials.api_fields_read


class JobCredentialsList(BaseResponse):
    # NOTE: 'global' attribute not usable..
    shared = serializers.ListSerializer(child=JobGlobalCredentialsReadResponse())
    user = serializers.ListSerializer(child=JobUserCredentialsReadResponse())


class JobGlobalCredentialsWriteRequest(serializers.ModelSerializer):
    class Meta:
        model = JobGlobalCredentials
        fields = JobGlobalCredentials.api_fields_write

    vault_pass = serializers.CharField(max_length=100, required=False, default=None, allow_blank=True)
    become_pass = serializers.CharField(max_length=100, required=False, default=None, allow_blank=True)
    connect_pass = serializers.CharField(max_length=100, required=False, default=None, allow_blank=True)
    ssh_key = serializers.CharField(max_length=2000, required=False, default=None, allow_blank=True)


class JobUserCredentialsWriteRequest(JobGlobalCredentialsWriteRequest):
    class Meta:
        model = JobUserCredentials
        fields = JobUserCredentials.api_fields_write


def are_global_credentials(request) -> bool:
    if 'global' in request.GET and request.GET['global'] != 'true':
        return False

    return True


def _find_credentials(
        credentials_id: int, are_global: bool, user: USERS
) -> (BaseJobCredentials, None):
    try:
        if are_global:
            return JobGlobalCredentials.objects.get(id=credentials_id)

        return JobUserCredentials.objects.get(id=credentials_id, user=user)

    except ObjectDoesNotExist:
        return None


def _log_global_user(are_global: bool, lower: bool = False) -> str:
    if are_global:
        msg = 'Global credentials'

    else:
        msg = 'User credentials'

    if lower:
        return msg.lower()

    return msg


def credentials_in_use(credentials: BaseJobCredentials) -> bool:
    if isinstance(credentials, JobGlobalCredentials):
        in_use_jobs = Job.objects.filter(credentials_default=credentials).exists()
        in_use_execs = JobExecution.objects.filter(credential_global=credentials).exists()
        in_use = in_use_jobs or in_use_execs

    else:
        in_use = JobExecution.objects.filter(credential_user=credentials).exists()

    return in_use


SSH_KEY_PREFIX = '-----BEGIN OPENSSH PRIVATE KEY-----'
SSH_KEY_APPENDIX = '-----END OPENSSH PRIVATE KEY-----'


def _validate_and_fix_ssh_key(key: str) -> (str, None):
    if key.find(SSH_KEY_PREFIX) == -1:
        # only support unencrypted keys for now
        return None

    key = key.replace(SSH_KEY_PREFIX, '').replace(SSH_KEY_APPENDIX, '').strip().replace(' ', '\n')
    return f'{SSH_KEY_PREFIX}\n{key}\n{SSH_KEY_APPENDIX}\n'


class APIJobCredentials(APIView):
    http_method_names = ['get', 'post']
    serializer_class = GenericResponse
    permission_classes = API_PERMISSION
    parameters = [
        OpenApiParameter(
            name='global', type=bool, default=True,
            description='If the credentials are global or user-specific',
            required=False,
        ),
    ]

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(JobCredentialsList, description='Return list of credentials'),
        },
        summary='Return list of all credentials the current user is privileged to view.',
        operation_id='credentials_list',
    )
    def get(self, request):
        user = get_api_user(request)
        credentials_global = []
        credentials_global_raw = JobGlobalCredentials.objects.all()
        for credentials in credentials_global_raw:
            if has_credentials_permission(
                    user=user,
                    credentials=credentials,
                    permission_needed=CHOICE_PERMISSION_READ,
            ):
                credentials_global.append(JobGlobalCredentialsReadResponse(instance=credentials).data)

        credentials_user_raw = JobUserCredentials.objects.filter(user=user)
        credentials_user = []
        for credentials in credentials_user_raw:
            credentials_user.append(JobUserCredentialsReadResponse(instance=credentials).data)

        return Response(data={'shared': credentials_global, 'user': credentials_user}, status=200)

    @extend_schema(
        request=JobGlobalCredentialsWriteRequest,
        responses={
            200: OpenApiResponse(GenericResponse, description='Credentials created'),
            400: OpenApiResponse(GenericResponse, description='Invalid credentials data provided'),
            403: OpenApiResponse(GenericResponse, description='Not privileged to create global credentials'),
        },
        summary='Create credentials.',
        operation_id='credentials_create',
        parameters=parameters,
    )
    def post(self, request):
        user = get_api_user(request)
        are_global = are_global_credentials(request)

        if are_global and not has_manager_privileges(user=user, kind='credentials'):
            return Response(data={'msg': 'Not privileged to create global credentials'}, status=403)

        if are_global:
            self.serializer_class = JobGlobalCredentialsWriteRequest

        else:
            self.serializer_class = JobUserCredentialsWriteRequest

        serializer = self.serializer_class(data=request.data)

        if not serializer.is_valid():
            return Response(
                data={'msg': f"Provided {_log_global_user(are_global)} data is not valid: '{serializer.errors}'"},
                status=400,
            )

        for field in BaseJobCredentials.SECRET_ATTRS:
            value = serializer.validated_data[field]
            if field in BaseJobCredentials.SECRET_ATTRS:
                if is_null(value) or value == BaseJobCredentials.SECRET_HIDDEN:
                    serializer.validated_data[field] = None

                elif field == 'ssh_key':
                    value = _validate_and_fix_ssh_key(value)
                    if value is None:
                        return Response(
                            data={'msg': f"Provided {_log_global_user(are_global)} ssh-key is not valid'"},
                            status=400,
                        )

                    serializer.validated_data[field] = value

        try:
            serializer.save()

        except IntegrityError as err:
            return Response(
                data={'msg': f"Provided {_log_global_user(are_global)} data is not valid: '{err}'"},
                status=400,
            )

        return Response(data={'msg': f'{_log_global_user(are_global, lower=False)} created'}, status=200)


class APIJobCredentialsItem(APIView):
    http_method_names = ['get', 'delete', 'put']
    serializer_class = GenericResponse
    permission_classes = API_PERMISSION
    parameters = [
        OpenApiParameter(
            name='global', type=bool, default=True,
            description='If the credentials are global or user-specific',
            required=False,
        ),
    ]

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(JobUserCredentialsReadResponse, description='Return information about credentials'),
            403: OpenApiResponse(GenericResponse, description='Not privileged to view the credentials'),
            404: OpenApiResponse(GenericResponse, description='Credentials not exist'),
        },
        summary='Return information about a set of credentials.',
        operation_id='credentials_view',
        parameters=parameters,
    )
    def get(self, request, credentials_id: int):
        user = get_api_user(request)
        are_global = are_global_credentials(request)

        if are_global:
            self.serializer_class = JobGlobalCredentialsReadResponse

        else:
            self.serializer_class = JobUserCredentialsReadResponse

        credentials = _find_credentials(
            credentials_id=credentials_id,
            are_global=are_global,
            user=user,
        )
        if credentials is None:
            base_msg = f"{_log_global_user(are_global)} with ID {credentials_id} do not exist"
            if not are_global:
                return Response(data={'msg': f"{base_msg} or belong to another user"}, status=404)

            return Response(data={'msg': base_msg}, status=404)

        if are_global and not has_credentials_permission(
                user=user,
                credentials=credentials,
                permission_needed=CHOICE_PERMISSION_READ,
        ):
            return Response(
                data={'msg': f"{_log_global_user(are_global)} '{credentials.name}' are not viewable"},
                status=403,
            )

        return Response(data=self.serializer_class(instance=credentials).data, status=200)

    @extend_schema(
        request=None,
        responses={
            200: OpenApiResponse(GenericResponse, description='Credentials deleted'),
            403: OpenApiResponse(GenericResponse, description='Not privileged to delete the credentials'),
            404: OpenApiResponse(GenericResponse, description='Credentials do not exist'),
        },
        summary='Delete credentials.',
        operation_id='credentials_delete',
        parameters=parameters,
    )
    def delete(self, request, credentials_id: int):
        user = get_api_user(request)
        are_global = are_global_credentials(request)

        if are_global:
            self.serializer_class = JobGlobalCredentialsReadResponse

        else:
            self.serializer_class = JobUserCredentialsReadResponse

        credentials = _find_credentials(
            credentials_id=credentials_id,
            are_global=are_global,
            user=user,
        )
        if credentials is None:
            base_msg = f"{_log_global_user(are_global)} with ID {credentials_id} do not exist"
            if not are_global:
                return Response(data={'msg': f"{base_msg} or belong to another user"}, status=404)

            return Response(data={'msg': base_msg}, status=404)

        if are_global and not has_credentials_permission(
                user=user,
                credentials=credentials,
                permission_needed=CHOICE_PERMISSION_DELETE,
        ):
            return Response(
                data={'msg': f"Not privileged to delete the {_log_global_user(are_global, lower=True)} "
                             f"'{credentials.name}'"},
                status=403)

        if credentials_in_use(credentials):
            return Response(
                data={'msg': f"{_log_global_user(are_global)} '{credentials.name}' cannot be deleted "
                             "as they are still in use"},
                status=400,
            )

        credentials.delete()
        return Response(data={'msg': f"{_log_global_user(are_global)} '{credentials.name}' deleted"}, status=200)

    @extend_schema(
        request=JobGlobalCredentialsWriteRequest,
        responses={
            200: OpenApiResponse(GenericResponse, description='Credentials updated'),
            400: OpenApiResponse(GenericResponse, description='Invalid credentials data provided'),
            403: OpenApiResponse(GenericResponse, description='Not privileged to modify the credentials'),
            404: OpenApiResponse(GenericResponse, description='Credentials do not exist'),
        },
        summary='Modify credentials.',
        operation_id='credentials_edit',
        parameters=parameters,
    )
    def put(self, request, credentials_id: int):
        user = get_api_user(request)
        are_global = are_global_credentials(request)

        if are_global:
            self.serializer_class = JobGlobalCredentialsWriteRequest

        else:
            self.serializer_class = JobUserCredentialsWriteRequest

        credentials = _find_credentials(
            credentials_id=credentials_id,
            are_global=are_global,
            user=user,
        )
        if credentials is None:
            base_msg = f"{_log_global_user(are_global)} with ID {credentials_id} do not exist"
            if not are_global:
                return Response(data={'msg': f"{base_msg} or belong to another user"}, status=404)

            return Response(data={'msg': base_msg}, status=404)

        if are_global and not has_credentials_permission(
                user=user,
                credentials=credentials,
                permission_needed=CHOICE_PERMISSION_WRITE,
        ):
            return Response(
                data={'msg': f"Not privileged to modify the {_log_global_user(are_global, lower=True)} "
                             f"'{credentials.name}'"},
                status=403,
            )

        serializer = self.serializer_class(data=request.data)
        if not serializer.is_valid():
            return Response(
                data={'msg': f"Provided {_log_global_user(are_global, lower=True)} data is not valid: "
                             f"'{serializer.errors}'"},
                status=400,
            )

        try:
            # not working with password properties: 'Job.objects.filter(id=job_id).update(**serializer.data)'
            for field, value in serializer.data.items():
                if field in BaseJobCredentials.SECRET_ATTRS:
                    if is_null(value) or value == BaseJobCredentials.SECRET_HIDDEN:
                        value = getattr(credentials, field)

                    elif field == 'ssh_key':
                        value = _validate_and_fix_ssh_key(value)
                        if value is None:
                            return Response(
                                data={'msg': f"Provided {_log_global_user(are_global)} ssh-key is not valid'"},
                                status=400,
                            )

                elif field == 'user':
                    continue

                setattr(credentials, field, value)

            if not are_global:
                credentials.user = user

            credentials.save()

        except IntegrityError as err:
            return Response(
                data={'msg': f"Provided {_log_global_user(are_global, lower=True)} data is not valid: '{err}'"},
                status=400,
            )

        return Response(data={'msg': f"{_log_global_user(are_global)} '{credentials.name}' updated"}, status=200)



# pylint: disable=E0401
from django.urls import path, re_path
from django.conf.urls import include
from django.contrib import admin
from django.contrib.auth.views import LoginView, PasswordChangeView

from web_serve_static import urlpatterns_static
from aw.api import urlpatterns_api
from aw.views.main import urlpatterns_ui, catchall, logout
from aw.config.environment import check_aw_env_var_true, auth_mode_saml
from aw.utils.deployment import deployment_dev
from aw.views.forms.auth import saml_sp_initiated_login, saml_sp_initiated_login_init

urlpatterns = []

if deployment_dev() or check_aw_env_var_true(var='serve_static', fallback=True):
    urlpatterns += urlpatterns_static

urlpatterns += urlpatterns_api

# AUTH
if auth_mode_saml():
    urlpatterns += [
        path('a/saml/init/', saml_sp_initiated_login_init),
        path('a/saml/', include('django_saml2_auth.urls')),
        # user views
        path('a/login/', saml_sp_initiated_login, name='login_sso'),
        path('a/login/fallback/', LoginView.as_view(), name='login'),
    ]

else:
    urlpatterns += [
        path('a/login/', LoginView.as_view(), name='login'),
    ]

urlpatterns += [
    path('a/password_change/', PasswordChangeView.as_view()),
    path('_admin/', admin.site.urls),
    path('o/', logout),
]

# UI
urlpatterns += urlpatterns_ui
urlpatterns += [
    # fallback
    re_path(r'^', catchall),
]
